# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot battery-relevant curves,
such as voltages and volume expansions.

"""

from typing import List, Optional, Dict, Union

import numpy as np
from matador.utils.chem_utils import (
    get_formula_from_stoich,
    get_subscripted_formula_tex,
)
from matador.plotting.plotting import plotting_function, SAVE_EXTS
from matador.battery import VoltageProfile

EPS = 1e-12


__all__ = ["plot_voltage_curve", "plot_volume_curve"]


@plotting_function
def plot_voltage_curve(
    profiles: Union[List[VoltageProfile], VoltageProfile],
    ax=None,
    show: bool = False,
    labels: bool = False,
    savefig: Optional[str] = None,
    curve_labels: Optional[Union[str, List[str]]] = None,
    line_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    shade_average_voltage: bool = False,
    colour_by_stoichiometry: bool = False,
    plot_average_voltage: bool = False,
    plot_max_capacity: bool = False,
    legend: bool = True,
    label_suffix: str = None,
    expt: Optional[str] = None,
    expt_label: Optional[str] = None,
):
    """Plot voltage curve calculated for phase diagram.

    Parameters:
        profiles (list/VoltageProfile): list of/single voltage profile(s).

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        show (bool): whether to show plot in an X window.
        savefig (str): filename to use to save the plot.
        curve_labels (list): optional list of labels for the curves in
            the profiles list.
        line_kwargs (list or dict): parameters to pass to the curve plotter,
            if a list then the line kwargs will be passed to each line individually.
        shade_average_voltage (bool): whether to shade the region spanned from average
            voltage to max capacity.
        plot_average_voltage (bool): whether to draw a guideline at the average voltage.
        plot_max_capacity (bool): whether to draw a guideline at the maximum capacity.
        legend (bool): whether to draw the legend.
        label_suffix (str): A suffix to add to every label on the plot (useful when combining axes)
        colour_by_stoichiometry (bool): whether to use mixed VESTA colours for each curve
            based on the starting formula of the electrode.
        expt (str): string to a filename of a csv Q, V to add to the plot.
        expt_label (str): label for any experimental profile passed to the plot.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if not isinstance(profiles, list):
        profiles = [profiles]
    if curve_labels is not None and not isinstance(curve_labels, list):
        curve_labels = [curve_labels]
    if line_kwargs is not None and not isinstance(line_kwargs, list):
        line_kwargs = len(profiles) * [line_kwargs]

    if curve_labels is not None and len(curve_labels) != len(profiles):
        raise RuntimeError(
            "Wrong number of labels passed for number of profiles: {} vs {}".format(
                len(curve_labels), len(profiles)
            )
        )

    if line_kwargs is not None and len(line_kwargs) != len(profiles):
        raise RuntimeError(
            "Wrong number of line kwargs passed for number of profiles: {} vs {}".format(
                len(line_kwargs), len(profiles)
            )
        )

    dft_label = None
    if expt is not None:
        expt_data = np.loadtxt(expt, delimiter=",")
        if expt_label:
            ax.plot(
                expt_data[:, 0], expt_data[:, 1], c="k", lw=2, ls="-", label=expt_label
            )
        else:
            ax.plot(
                expt_data[:, 0],
                expt_data[:, 1],
                c="k",
                lw=2,
                ls="-",
                label="Experiment",
            )

        if len(profiles) == 1:
            dft_label = "DFT (this work)"

    line_colours = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])

    for ind, profile in enumerate(profiles):
        if dft_label is None and curve_labels is None:
            stoich_label = get_formula_from_stoich(
                profile.starting_stoichiometry, tex=True
            )
        else:
            stoich_label = None

        _label = stoich_label if dft_label is None else dft_label
        if curve_labels is not None and len(curve_labels) > ind:
            _label = curve_labels[ind]

        if label_suffix:
            _label += f" {label_suffix}"

        _line_kwargs = {"c": line_colours[(ind % len(line_colours)) + 2]}
        if line_kwargs is not None:
            _line_kwargs.update(line_kwargs[ind])

        if colour_by_stoichiometry:
            from matador.utils.viz_utils import formula_to_colour

            colour = formula_to_colour(
                get_formula_from_stoich(profile.starting_stoichiometry)
            )
            _line_kwargs["c"] = colour
        else:
            colour = _line_kwargs["c"]

        _add_voltage_curve(
            profile.capacities, profile.voltages, ax, label=_label, **_line_kwargs
        )

        if shade_average_voltage:
            ax.fill_between(
                np.linspace(
                    0,
                    np.nanmax(profile.capacities),
                    2,
                ),
                np.linspace(0, 0, 2),
                [profile.average_voltage, profile.average_voltage],
                lw=1,
                edgecolor=colour,
                facecolor=colour + "20",
                zorder=-ind + 2,
                ls="--",
            )

        if plot_average_voltage:
            vline_kwargs = _line_kwargs
            vline_kwargs["ls"] = "--"
            vline_kwargs["alpha"] = 0.8
            vline_kwargs["zorder"] = 0
            ax.axhline(profile.average_voltage, **vline_kwargs)

        if plot_max_capacity:
            vline_kwargs = _line_kwargs
            vline_kwargs["ls"] = "-."
            vline_kwargs["alpha"] = 0.8
            vline_kwargs["zorder"] = 0
            ax.axvline(np.nanmax(profile.capacities), **vline_kwargs)

    if labels:
        if len(profiles) > 1:
            print("Only labelling first voltage profile.")
        for ind, reaction in enumerate(profiles[0].reactions[:-1]):
            _labels = []
            for phase in reaction:
                if phase[0] is None or phase[0] == 1.0:
                    _label = ""
                else:
                    _label = "{:.1f} ".format(phase[0])
                _label += "{}".format(get_subscripted_formula_tex(phase[1]))
                _labels.append(_label)
            _label = "+".join(_labels)
            _position = (
                profiles[0].capacities[ind + 1],
                profiles[0].voltages[ind + 1] + max(profiles[0].voltages) * 0.01,
            )
            ax.annotate(
                _label, xy=_position, textcoords="data", ha="center", zorder=9999
            )

    if expt or len(profiles) > 1:

        if plot_average_voltage:
            ax.axvline(-1, c="k", ls="--", alpha=0.8, label="Average voltages")
        if plot_max_capacity:
            ax.axvline(
                -1, c="k", ls="-.", alpha=0.8, label="Maximum gravimetric capacity"
            )

        if legend:
            ax.legend()

    ax.set_ylabel("Voltage (V) vs {ion}$^+/${ion}".format(ion=profile.active_ion))
    ax.set_xlabel("Gravimetric cap. (mAh/g)")

    _, end = ax.get_ylim()
    from matplotlib.ticker import MultipleLocator

    ax.yaxis.set_major_locator(MultipleLocator(0.2))
    ax.set_ylim(0, 1.1 * end)
    _, end = ax.get_xlim()
    ax.set_xlim(0, 1.1 * end)
    ax.grid(False)
    plt.tight_layout(pad=0.0, h_pad=1.0, w_pad=0.2)

    if savefig:
        plt.savefig(savefig)
        print("Wrote {}".format(savefig))

    elif show:
        plt.show()

    return ax


def _add_voltage_curve(capacities, voltages, ax, label=None, **kwargs):
    """Add the voltage curves stored under hull['voltage_data'] to the plot.

    Parameters:
        capacities (list): list or numpy array of capacities.
        voltages (list): list or numpy array of voltages.
        ax (matplotlib.axes.Axes): an existing axis object on which to plot.

    Keyword arguments:
        **kwargs (dict): to pass to matplotlib, using abbreviated names (e.g. 'c' not 'color')
        label (str): if present, add this label to the first segment of the voltage curve
            so it can be added to the legend.
        alpha (float): transparency of line

    """
    line_kwargs = {
        "lw": 2,
        "alpha": 1,
    }
    line_kwargs.update(kwargs)

    for i in range(1, len(voltages) - 1):
        if i == 1 and label is not None:
            ax.plot(
                [capacities[i - 1], capacities[i]],
                [voltages[i], voltages[i]],
                label=label,
                **line_kwargs,
            )
        else:
            ax.plot(
                [capacities[i - 1], capacities[i]],
                [voltages[i], voltages[i]],
                **line_kwargs,
            )
        if i != len(voltages) - 2:
            ax.plot(
                [capacities[i], capacities[i]],
                [voltages[i], voltages[i + 1]],
                **line_kwargs,
            )


@plotting_function
def plot_volume_curve(
    hull,
    ax=None,
    show=True,
    legend=False,
    as_percentages=False,
    label=None,
    label_suffix=None,
    exclude_elemental=False,
    line_kwargs: Optional[Union[Dict, List[Dict]]] = None,
    colour_by_stoichiometry: bool = False,
    end_marker_only: bool = True,
    **kwargs,
):
    """Plot volume curve calculated for phase diagram.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot.
        show (bool): whether or not to display plot in X-window.
        legend (bool): whether to add the legend.
        label (str/list): Either a list of labels or a single label for the volume curves.
        label_suffix (str): A suffix to add to every label on the plot (useful when combining axes)
        as_percentages (bool): whether to show the expansion as a percentage.
        exclude_elemental (bool): whether to include any elemental electrodes
            when considering ternary phase diagrams.
        line_kwargs (list or dict): parameters to pass to the curve plotter,
            if a list then the line kwargs will be passed to each line individually.
        colour_by_stoichiometry (bool): whether to use mixed VESTA colours for each curve
            based on the starting formula of the electrode.
        end_marker_only (bool): Only place a marker on the worst volume expansion for each curve.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        if hull.savefig or any([kwargs.get(ext) for ext in SAVE_EXTS]):
            fig = plt.figure()
        else:
            fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        ax = ax

    if label and not isinstance(label, list):
        label = [label]

    num_curves = len(hull.volume_data["Q"])

    if line_kwargs is not None and not isinstance(line_kwargs, list):
        line_kwargs = num_curves * [line_kwargs]

    if as_percentages:
        volume_key = "volume_expansion_percentage"
    else:
        volume_key = "volume_ratio_with_bulk"

    line_colours = plt.rcParams["axes.prop_cycle"].by_key()["color"]

    for j in range(num_curves):
        if exclude_elemental and len(hull.volume_data["endstoichs"][j]) == 1:
            continue

        _line_kwargs = {"c": line_colours[(j % len(line_colours)) + 2]}
        if line_kwargs is not None:
            _line_kwargs.update(line_kwargs[j])

        if colour_by_stoichiometry:
            from matador.utils.viz_utils import formula_to_colour

            colour = formula_to_colour(
                get_formula_from_stoich(hull.volume_data["endstoichs"][j])
            )
            _line_kwargs["c"] = colour
        else:
            colour = _line_kwargs["c"]

        lw = _line_kwargs.pop("lw", 2)
        marker = _line_kwargs.pop("marker", "o")
        markeredgewidth = _line_kwargs.pop("markeredgewidth", 1.5)
        markeredgecolor = _line_kwargs.pop("markeredgecolor", "k")

        stable_hull_dist = hull.volume_data["hull_distances"][j]
        if len(stable_hull_dist) != len(hull.volume_data["Q"][j]):
            raise RuntimeError("This plot does not support --hull_cutoff.")

        if end_marker_only:
            ax.plot(
                [0, np.max(hull.volume_data["Q"][j])],
                [1, np.max(hull.volume_data[volume_key][j])],
                marker=marker,
                markeredgewidth=markeredgewidth,
                markeredgecolor=markeredgecolor,
                zorder=1000,
                lw=0,
                **_line_kwargs,
            )

        else:
            ax.plot(
                [
                    q
                    for ind, q in enumerate(hull.volume_data["Q"][j])
                    if stable_hull_dist[ind] <= EPS
                ],
                [
                    v
                    for ind, v in enumerate(hull.volume_data[volume_key][j])
                    if stable_hull_dist[ind] <= EPS
                ],
                marker=marker,
                markeredgewidth=markeredgewidth,
                markeredgecolor=markeredgecolor,
                zorder=1000,
                lw=0,
                **_line_kwargs,
            )

        if label:
            _label = label[j]
        else:
            _label = get_formula_from_stoich(
                hull.volume_data["endstoichs"][j], tex=True
            )
        if label_suffix:
            _label += f" {label_suffix}"

        ax.plot(
            [
                q
                for ind, q in enumerate(hull.volume_data["Q"][j])
                if stable_hull_dist[ind] <= EPS
            ],
            [
                v
                for ind, v in enumerate(hull.volume_data[volume_key][j])
                if stable_hull_dist[ind] <= EPS
            ],
            label=_label,
            lw=lw,
            **_line_kwargs,
        )

    ax.set_xlabel("Gravimetric capacity (mAh/g)")
    if as_percentages:
        ax.set_ylabel("Volume expansion (%)")
    else:
        if len(hull.volume_data["endstoichs"]) == 1:
            ax.set_ylabel(
                f"Volume ratio with {get_formula_from_stoich(hull.volume_data['endstoichs'][0], tex=True)} starting electrode"
            )
        else:
            ax.set_ylabel("Volume ratio with starting electrode")

    if legend or len(hull.volume_data["Q"]) > 1:
        ax.legend()

    return ax



""" This submodule includes functionality to create a
template .matadorrc file.

"""

__all__ = ["quickstart_settings"]

import os
from matador.utils.print_utils import print_warning, print_success


def quickstart_settings():
    """Call this if no user-specified configuration file is found. A series
    of questions will be asked to create a basic matadorrc.

    """
    print(
        "No user-specified input found at either ~/.matadorrc or " "~/.config/matadorrc"
    )
    response = input("Would you like to use quickstart to create one now? [y/n] ")
    if not response.lower().startswith("y"):
        return None

    hline = 80 * "-"
    print("Okay, let's get started!")
    print("Firstly, where would you like me to put your config file: ")
    valid = False
    while not valid:
        response = input("[1] ~/.matadorrc, [2] ~/.config/matadorrc? ")
        if response not in ["1", "2"]:
            print("I didn't understand, sorry... please enter 1 or 2")
        else:
            valid = True

    if response == "1":
        fname = os.path.expanduser("~/.matadorrc")
    else:
        fname = os.path.expanduser("~/.config/matadorrc")

    flines = []
    print(hline)

    response = input(
        "Great! Would you like to use matador with a MongoDB database? [y/n] "
    )
    if response.lower().startswith("y"):
        flines.append("mongo:")
        host_response = input("Where is this MongoDB hosted? [default: localhost] ")
        if host_response == "":
            host_response = "localhost"
        flines.append("  host: {}".format(host_response))
        print(hline)
        port_response = input("and at what port? [default: 27017] ")
        if port_response == "":
            port_response = 27017
        flines.append("  port: {}".format(port_response))
        print(hline)
        print(
            "Please set your firewall and MongoDB settings accordingly: "
            "by default MongoDB\nneeds no authentication so be careful "
            "if your database is running on an open port!"
        )
        test_conn_response = input(
            "Would you like me to test the connection to this MongoDB instance now? [y/n] "
        )
        keep_trying = True
        if test_conn_response.lower() == "y":
            while keep_trying:
                import pymongo as pm

                cli = pm.MongoClient(
                    host_response,
                    port=int(port_response),
                    maxIdleTimeMS=600000,
                    socketTimeoutMS=3600000,
                    serverSelectionTimeoutMS=10000,
                    connectTimeoutMS=10000,
                )
                try:
                    cli.database_names()
                    keep_trying = False
                except pm.errors.ServerSelectionTimeoutError:
                    print_warning(
                        "Failed to connect to {}:{}, are you sure it is running?".format(
                            host_response, port_response
                        )
                    )
                    response = input("Would you like me to try again? [y/n] ")
                    keep_trying = response.lower().startswith("y")
                except Exception:
                    keep_trying = False

        print(hline)
        db_response = input(
            "What is the name of the database (not collection) "
            "that you want to query/create? [default: crystals] "
        )
        if db_response == "":
            db_response = "crystals"
        flines.append("  db: {}".format(db_response))

        print(hline)
        coll_response = input(
            "What about the default collection name to query? [default: repo] "
        )
        if coll_response == "":
            coll_response = "repo"
        flines.append("  default_collection: {}".format(coll_response))

        print(hline)
        print("Would you like to set a protected folder for this collection?")
        file_path_response = input(
            "Any import to this collection must happen from this path, and the "
            "--force flag must be used if so. [default: None] "
        )
        if file_path_response != "":
            flines.append(
                "  default_file_collection_path: {}".format(file_path_response)
            )

        print(hline)
        print_success(
            "Mongo section of matadorrc complete! Saving as {} and exiting...".format(
                fname
            )
        )
        print(hline)

    if not flines:
        print(
            "Okay, there is no config file to be made and I will keep using the defaults."
        )
        return None

    with open(fname, "a") as f:
        for line in flines:
            f.write(line + "\n")

    return fname



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule provides data loaders for the tabulated electric
quadrupole moment data stored within the module.

"""


def _load_quadrupole_moments_from_file():
    """Load the electric quadrupole moment file into the constant dictionary
    ELECTRIC_QUADRUPOLE_MOMENTS.

    """

    from pathlib import Path
    import numpy as np
    from matador.data.periodic_table import PERIODIC_TABLE

    filename = Path(__file__).parent.joinpath("electric_quadrupole_moments.csv")
    moments = np.loadtxt(filename, comments="#", delimiter=",")
    moments_dict = {}
    for el, col in zip(PERIODIC_TABLE.values(), moments):
        moments_dict[el.symbol] = moments[el.number - 1]

    return moments_dict


ELECTRIC_QUADRUPOLE_MOMENTS = _load_quadrupole_moments_from_file()



import unittest
import numpy as np


class VizUtilTest(unittest.TestCase):
    """Tests viz util functions."""

    def test_formula_to_colour(self):
        from matador.utils.viz_utils import formula_to_colour

        self.assertListEqual(formula_to_colour("K"), [0.4, 0.14, 0.43, 1])
        self.assertListEqual(formula_to_colour("KP"), [0.62, 0.275, 0.255, 1])
        self.assertListEqual(formula_to_colour("P"), [0.84, 0.41, 0.08, 1])
        np.testing.assert_array_almost_equal(
            formula_to_colour("KSnP"), [0.60666, 0.366666, 0.3999999, 1], decimal=3
        )
        with self.assertRaises(RuntimeError):
            formula_to_colour("KSnPSb")



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the :class:`CastepPhononWorkflow`
class, which performs phonon calculations with CASTEP in
multiple steps (only when necessary):

    1. Try to pre-relax structure (skipped if check file
           is already present).
    2. Calculate dynamical matrix.
    3. If ``phonon_fine_kpoint_mp_spacing`` keyword is found,
       interpolate dynamical matrix to form phonon DOS.
    4. If ``phonon_fine_kpoint_path_spacing`` key word is found,
       interpolate dynamical matrix to form phonon dispersion
       on the path given by seekpath.
    5. If ``task=thermodynamics``, perform a CASTEP thermodynamics
       calculation for the temperature-dependence of the free energy.

"""


import os
import copy
import logging
from matador.workflows.workflows import Workflow

LOG = logging.getLogger("run3")


def castep_full_phonon(computer, calc_doc, seed, **kwargs):
    """Perform a "full" phonon calculation on a system, i.e.
    first perform a relaxation in a standardised unit cell,
    then compute the dynamical matrix, then finally interpolate
    that dynamical matrix into dispersion curves and DOS. This function
    is a wrapper for the :class:`CastepPhononWorkflow` class.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepPhononWorkflow(computer, calc_doc, seed, **kwargs)
    return workflow.success


class CastepPhononWorkflow(Workflow):
    """Perform a "full" phonon calculation on a system, i.e.
    first perform a relaxation in a standardised unit cell,
    then compute the dynamical matrix, then finally interpolate
    that dynamical matrix into dispersion curves and DOS.

    Attributes:
        computer (:obj:`ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.

    """

    def preprocess(self):
        """Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """
        # default todo
        todo = {
            "relax": True,
            "dynmat": True,
            "vdos": False,
            "dispersion": False,
            "thermodynamics": False,
        }
        # definition of steps and names
        steps = {
            "relax": castep_phonon_prerelax,
            "dynmat": castep_phonon_dynmat,
            "vdos": castep_phonon_dos,
            "dispersion": castep_phonon_dispersion,
            "thermodynamics": castep_phonon_thermodynamics,
        }

        exts = {
            "relax": {
                "input": [".cell", ".param"],
                "output": [".castep", "-out.cell", ".*err"],
            },
            "dynmat": {"input": [".cell", ".param"], "output": [".castep", ".*err"]},
            "vdos": {
                "input": [".cell", ".param"],
                "output": [".castep", ".phonon", ".phonon_dos", ".*err"],
            },
            "dispersion": {
                "input": [".cell", ".param"],
                "output": [".castep", ".phonon", ".*err"],
            },
            "thermodynamics": {
                "input": [".cell", ".param"],
                "output": [".castep", ".*err"],
            },
        }

        if self.calc_doc.get("task").lower() in [
            "phonon",
            "thermodynamics",
            "phonon+efield",
        ]:
            if (
                "phonon_fine_kpoint_path" in self.calc_doc
                or "phonon_fine_kpoint_list" in self.calc_doc
                or "phonon_fine_kpoint_path_spacing" in self.calc_doc
            ):
                todo["dispersion"] = True
            if "phonon_fine_kpoint_mp_spacing" in self.calc_doc:
                todo["vdos"] = True
            if self.calc_doc["task"].lower() == "thermodynamics":
                todo["thermodynamics"] = True

        # prepare to do pre-relax if there's no check file
        if os.path.isfile(self.seed + ".check"):
            todo["relax"] = False
            LOG.info(
                "Restarting from {}.check, so not performing re-relaxation".format(
                    self.seed
                )
            )

        for key in todo:
            if todo[key]:
                self.add_step(
                    steps[key],
                    key,
                    input_exts=exts[key].get("input"),
                    output_exts=exts[key].get("output"),
                )

        # always standardise the cell so that any phonon calculation can have
        # post-processing performed after the fact, unless a path has been provided
        if (
            "phonon_fine_kpoint_list" not in self.calc_doc
            and "phonon_fine_kpoint_path" not in self.calc_doc
        ):
            from matador.utils.cell_utils import cart2abc

            prim_doc, kpt_path = self.computer.get_seekpath_compliant_input(
                self.calc_doc,
                self.calc_doc.get("phonon_fine_kpoint_path_spacing", 0.02),
            )
            self.calc_doc.update(prim_doc)
            self.calc_doc["lattice_abc"] = cart2abc(self.calc_doc["lattice_cart"])
            if todo["dispersion"]:
                self.calc_doc["phonon_fine_kpoint_list"] = kpt_path

        elif todo["dispersion"] and "phonon_fine_kpoint_path" in self.calc_doc:
            self._user_defined_kpt_path = True
            LOG.warning("Using user-defined k-point path for all structures.")
            self.calc_doc["phonon_fine_kpoint_spacing"] = self.calc_doc.get(
                "phonon_fine_kpoint_path_spacing", 0.05
            )

        # always shift phonon grid to include Gamma
        if "phonon_kpoint_mp_spacing" in self.calc_doc:
            from matador.utils.cell_utils import calc_mp_grid, shift_to_include_gamma

            grid = calc_mp_grid(
                self.calc_doc["lattice_cart"], self.calc_doc["phonon_kpoint_mp_spacing"]
            )
            offset = shift_to_include_gamma(grid)
            if offset != [0, 0, 0]:
                self.calc_doc["phonon_kpoint_mp_offset"] = offset
                LOG.debug("Set phonon MP grid offset to {}".format(offset))

        LOG.info("Preprocessing completed: run3 phonon options {}".format(todo))


def castep_phonon_prerelax(computer, calc_doc, seed):
    """Run a singleshot geometry optimisation before an SCF-style calculation.
    This is typically used to ensure phonon calculations start successfully.
    The phonon calculation will then be restarted from the .check file produced here.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    from matador.workflows.castep.common import castep_prerelax

    LOG.info("Performing CASTEP phonon pre-relax...")
    required = ["write_checkpoint"]
    forbidden = [
        "phonon_fine_kpoint_list",
        "phonon_fine_kpoint_path",
        "phonon_fine_kpoint_mp_spacing",
        "phonon_fine_kpoint_path_spacing",
    ]

    return castep_prerelax(
        computer, calc_doc, seed, required_keys=required, forbidden_keys=forbidden
    )


def castep_phonon_dynmat(computer, calc_doc, seed):
    """Runs a singleshot phonon dynmat calculation, with no "fine_method" interpolation.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP dynmat calculation...")
    dynmat_doc = copy.deepcopy(calc_doc)
    dynmat_doc["write_checkpoint"] = "ALL"
    if calc_doc["task"].lower() == "phonon+efield":
        dynmat_doc["task"] = "phonon+efield"
    else:
        dynmat_doc["task"] = "phonon"

    dynmat_doc["continuation"] = "default"

    required = ["continuation", "write_checkpoint"]
    forbidden = [
        "phonon_fine_kpoint_list",
        "phonon_fine_kpoint_path",
        "phonon_fine_kpoint_mp_spacing",
        "phonon_fine_kpoint_path_spacing",
    ]

    computer.validate_calc_doc(dynmat_doc, required, forbidden)
    return computer.run_castep_singleshot(
        dynmat_doc, seed, keep=True, intermediate=True
    )


def castep_phonon_dos(computer, calc_doc, seed):
    """Runs a DOS interpolation on top of a completed
    phonon calculation.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP phonon DOS calculation...")
    dos_doc = copy.deepcopy(calc_doc)
    dos_doc["task"] = "phonon"
    dos_doc["phonon_calculate_dos"] = True
    dos_doc["continuation"] = "default"

    required = ["phonon_fine_kpoint_mp_spacing"]
    forbidden = [
        "phonon_fine_kpoint_list",
        "phonon_fine_kpoint_path",
        "phonon_fine_kpoint_path_spacing",
    ]

    computer.validate_calc_doc(dos_doc, required, forbidden)

    return computer.run_castep_singleshot(dos_doc, seed, keep=True, intermediate=True)


def castep_phonon_dispersion(computer, calc_doc, seed):
    """Runs a dispersion interpolation on top of a completed
    phonon calculation.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP phonon dispersion calculation...")
    disp_doc = copy.deepcopy(calc_doc)
    disp_doc["task"] = "phonon"
    disp_doc["phonon_calculate_dos"] = False
    disp_doc["continuation"] = "default"

    required = []
    forbidden = ["phonon_fine_kpoint_mp_spacing"]

    computer.validate_calc_doc(disp_doc, required, forbidden)

    return computer.run_castep_singleshot(disp_doc, seed, keep=True, intermediate=True)


def castep_phonon_thermodynamics(computer, calc_doc, seed):
    """Runs a "thermodynamics" interpolation on top of a completed
    phonon calculation, using the phonon_fine_kpoint_mp_grid.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP thermodynamics calculation...")
    thermo_doc = copy.deepcopy(calc_doc)
    thermo_doc["continuation"] = "default"
    thermo_doc["task"] = "thermodynamics"
    thermo_doc["phonon_calculate_dos"] = False

    required = ["phonon_fine_kpoint_mp_spacing"]
    forbidden = [
        "phonon_fine_kpoint_list",
        "phonon_fine_kpoint_path",
        "phonon_fine_kpoint_path_spacing",
    ]

    computer.validate_calc_doc(thermo_doc, required, forbidden)

    return computer.run_castep_singleshot(
        thermo_doc, seed, keep=True, intermediate=True
    )



# coding: utf-8
# Distributed under the terms of the MIT License.

""" Some simple utilities for making DB connections. """


import pymongo as pm
from matador.config import load_custom_settings


def make_connection_to_collection(
    coll_names,
    check_collection=False,
    allow_changelog=False,
    mongo_settings=None,
    override=False,
    import_mode=False,
    quiet=True,
    debug=False,
):
    """Connect to database of choice.

    Parameters:
        coll_names (str): name of collection.

    Keyword Arguments:
        check_collection (bool): check whether collections exist (forces connection)
        allow_changelog (bool): allow queries to collections with names prefixed by __
        mongo_settings (dict): dict containing mongo and related config
        override (bool): don't ask for user input from stdin and assume all is well
        quiet (bool): don't print very much.

    Returns:
        client (MongoClient): the connection to the database
        db (Database): the database to query
        collections (dict): Collection objects indexed by name

    """

    if mongo_settings is None:
        settings = load_custom_settings(no_quickstart=override)
    else:
        settings = mongo_settings

    if not quiet:
        print("Trying to connect to {host}:{port}/{db}".format(**settings["mongo"]))

    client = pm.MongoClient(
        host=settings["mongo"]["host"],
        port=settings["mongo"]["port"],
        connect=False,
        maxIdleTimeMS=600000,  # disconnect after 10 minutes idle
        socketTimeoutMS=3600000,  # give up on database after 1 hr without results
        serverSelectionTimeoutMS=10000,  # give up on server after 2 seconds without results
        connectTimeoutMS=10000,
    )  # give up trying to connect to new database after 2 seconds

    try:
        database_names = client.list_database_names()
        if not quiet:
            print("Success!")
    except pm.errors.ServerSelectionTimeoutError as exc:
        print("{}: {}".format(type(exc).__name__, exc))
        raise SystemExit(
            "Unable to connect to {host}:{port}/{db}, exiting...".format(
                **settings["mongo"]
            )
        )

    if settings["mongo"]["db"] not in database_names:
        if override:
            response = "y"
        else:
            response = input(
                "Database {db} does not exist at {host}:{port}/{db}, "
                "would you like to create it? (y/n) ".format(**settings["mongo"])
            )

        if response.lower() != "y":
            raise SystemExit("Exiting...")
        else:
            print("Creating database {}".format(settings["mongo"]["db"]))

    db = client[settings["mongo"]["db"]]
    possible_collections = [
        name for name in db.list_collection_names() if not name.startswith("__")
    ]
    collections = dict()
    # allow lists of collections for backwards-compat, though normally
    # we only want to connect to one at a time
    if coll_names is not None:
        if not isinstance(coll_names, list):
            coll_names = [coll_names]

        if len(coll_names) > 1:
            raise NotImplementedError(
                "Querying multiple collections is no longer supported."
            )

        for collection in coll_names:

            if not allow_changelog:
                if collection.startswith("__"):
                    raise SystemExit(
                        "Queries to collections prefixed with __ are VERBOTEN!"
                    )

                if collection not in possible_collections:
                    options = fuzzy_collname_match(collection, possible_collections)
                    if not options and check_collection:
                        client.close()
                        raise SystemExit("Collection {} not found!".format(collection))
                    else:
                        print(
                            "Collection {} not found, did you mean one of these?".format(
                                collection
                            )
                        )
                        for ind, value in enumerate(options[:10]):
                            print("({}):\t{}".format(ind, value))
                        if check_collection:
                            try:
                                choice = int(input("Please enter your choice: "))
                                collection = options[choice]
                            except Exception:
                                raise SystemExit("Invalid choice. Exiting...")
                        elif import_mode:
                            if override:
                                choice = "y"
                            else:
                                choice = input(
                                    "Are you sure you want to make a new collection called {}? (y/n) ".format(
                                        collection
                                    )
                                )
                            if choice.lower() != "y" and choice.lower != "yes":
                                try:
                                    choice = int(
                                        input(
                                            "Then please enter your choice from above: "
                                        )
                                    )
                                    collection = options[choice]
                                except Exception:
                                    raise SystemExit("Invalid choice. Exiting...")

            collections[collection] = db[collection]
    else:
        default_collection = settings["mongo"]["default_collection"]
        if default_collection not in possible_collections:
            if check_collection:
                client.close()
                raise SystemExit(
                    "Default collection {} not found!".format(default_collection)
                )
            else:
                print("Creating new collection {}...".format(default_collection))
        collections["repo"] = db[default_collection]

    return client, db, collections


def fuzzy_collname_match(trial, targets):
    """Do a noddy fuzzy match for bits between punctuation, e.g.
    matthews_cool_database will search for matthews, cool and database
    in the known collection names.

    Parameters:
        trial (str): database search name.
        targets (list): list of existing database names.

    Returns:
        list: list of roughly matching collection names ordered
            by occurence of tokens.

    """
    split_chars = set(
        [char for char in trial if (not char.isalpha() and not char.isdigit())]
    )
    tokens = trial
    for char in split_chars:
        tokens = tokens.replace(char, " ")
    tokens = tokens.split()
    options = {}
    for token in tokens:
        for option in targets:
            if token in option:
                if option not in options:
                    options[option] = 0
                else:
                    options[option] += 1
    options_list = [
        item[0]
        for item in list(
            sorted(options.items(), reverse=True, key=lambda item: item[1])
        )
    ]
    return options_list



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot convex hulls and phase
diagrams generally.

"""


from collections import defaultdict
import numpy as np
from matador.utils.chem_utils import get_stoich_from_formula, get_formula_from_stoich
from matador.utils.viz_utils import (
    get_element_colours,
    colour_from_ternary_concentration,
)
from matador.plotting.plotting import plotting_function, get_linear_cmap, SAVE_EXTS

EPS = 1e-12

__all__ = [
    "plot_2d_hull",
    "plot_ternary_hull",
    "plot_ensemble_hull",
    "plot_temperature_hull",
]


def _get_hull_labels(hull, label_cutoff=None, num_species=None, exclude_edges=True):
    """Return list of structures to labels on phase diagram.

    Parameters:
        hull (matador.hull.QueryConvexHull): phase diagram to plot.

    Keyword arguments:
        num_species (int): structures containing this number of species
            will be labelled.
        exclude_edges (bool): ignore any structure that has zero of any
            chemical potential (i.e the edge of the hull).

    Returns:
        label_cursor (list(dict)): list of matador documents to label.

    """
    if label_cutoff is None:
        label_cutoff = 0.0
    if isinstance(label_cutoff, list) and len(label_cutoff) == 2:
        label_cutoff = sorted(label_cutoff)
        # first, only apply upper limit as we need to filter by stoich aftewards
        label_cursor = [
            doc for doc in hull.cursor if doc["hull_distance"] <= label_cutoff[1]
        ]
    else:
        if isinstance(label_cutoff, list):
            assert (
                len(label_cutoff) == 1
            ), "Incorrect number of label_cutoff values passed, should be 1 or 2."
            label_cutoff = label_cutoff[0]
        label_cursor = [
            doc for doc in hull.cursor if doc["hull_distance"] <= label_cutoff + EPS
        ]

    label_cursor = sorted(label_cursor, key=lambda _: _["hull_distance"], reverse=False)

    num_labels = len(
        {get_formula_from_stoich(doc["stoichiometry"]) for doc in label_cursor}
    )
    if num_labels < len(label_cursor):
        tmp_cursor = []
        for doc in label_cursor:
            if doc["stoichiometry"] not in [
                _doc["stoichiometry"] for _doc in tmp_cursor
            ]:
                tmp_cursor.append(doc)
            else:
                label_cursor = tmp_cursor
    if isinstance(label_cutoff, list) and len(label_cutoff) == 2:
        # now apply lower limit
        label_cursor = [
            doc
            for doc in label_cursor
            if label_cutoff[0] <= doc["hull_distance"] <= label_cutoff[1]
        ]
    # remove chemical potentials and unwanted e.g. binaries
    if num_species is not None:
        label_cursor = [
            doc for doc in label_cursor if len(doc["stoichiometry"]) == num_species
        ]
    if exclude_edges:
        label_cursor = [
            doc
            for doc in label_cursor
            if (all(doc["concentration"]) > 0 and sum(doc["concentration"]) <= 1 - EPS)
        ]

    label_cursor = sorted(label_cursor, key=lambda doc: doc["concentration"])

    return label_cursor


@plotting_function
def plot_2d_hull(
    hull,
    ax=None,
    show=True,
    plot_points=True,
    plot_tie_line=True,
    plot_hull_points=True,
    labels=None,
    label_cutoff=None,
    colour_by_source=False,
    sources=None,
    hull_label=None,
    source_labels=None,
    title=True,
    plot_fname=None,
    show_cbar=True,
    colour_by_composition=False,
    label_offset=(1, 0.1),
    specific_label_offset=None,
    eform_limits=None,
    legend_kwargs=None,
    hull_dist_unit="meV",
    **kwargs,
):
    """Plot calculated hull, returning ax and fig objects for further editing.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        ax (matplotlib.axes.Axes): an existing axis on which to plot,
        show (bool): whether or not to display the plot in an X window,
        plot_points (bool): whether or not to display off-hull structures,
        plot_hull_points (bool): whether or not to display on-hull structures,
        labels (bool): whether to label formulae of hull structures, also read from
            hull.args.
        label_cutoff (float/:obj:`tuple` of :obj:`float`): draw labels less than or
            between these distances form the hull, also read from hull.args.
        specific_label_offset (dict): A dictionary keyed by chemical formula strings
            containing 2-length float tuples of label offsets. A value of `None` for
            a particular formula will exclude that label from the plot.
        colour_by_source (bool): plot and label points by their sources
        show_cbar: Whether or not to plot the hull distance colourbar.
        alpha (float): alpha value of points when colour_by_source is True
        sources (list): list of possible provenances to colour when colour_by_source
            is True (others will be grey)
        title (str/bool): whether to include a plot title.
        png/pdf/svg (bool): whether or not to write the plot to a file.
        plot_fname (str): filename to write plot to, without file extension.

    Returns:
        matplotlib.axes.Axes: matplotlib axis with plot.

    """
    import matplotlib.pyplot as plt
    import matplotlib.colors

    # cache the specific label offset dict
    _specific_label_offset = None
    if specific_label_offset:
        _specific_label_offset = dict(specific_label_offset)

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if not hasattr(hull, "colours"):
        hull.colours = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])
    hull.default_cmap_list = get_linear_cmap(hull.colours[1:4], list_only=True)
    hull.default_cmap = get_linear_cmap(hull.colours[1:4], list_only=False)

    if colour_by_composition:
        try:
            element_colours = [get_element_colours()[s] for s in hull.species]
        except Exception:
            raise RuntimeError(
                f"Cannot `colour_by_composition`: no colour found for species {hull.species}."
            )
        conc_cmap = get_linear_cmap(element_colours, list_only=False)

    if labels is None:
        labels = hull.args.get("labels", False)
    if label_cutoff is None:
        label_cutoff = hull.args.get("label_cutoff")

    scale = 1
    scatter = []
    chempot_labels = [
        get_formula_from_stoich(get_stoich_from_formula(species, sort=False), tex=True)
        for species in hull.species
    ]
    tie_line = hull.convex_hull.points[hull.convex_hull.vertices]

    # plot hull structures
    if plot_hull_points:
        if colour_by_composition:
            hull_point_scale = 75
            edgewidth = 1.5
            edgecolor = kwargs.get("markeredgecolor", "black")
        else:
            hull_point_scale = 40
            edgewidth = 1.5
            edgecolor = kwargs.get("markeredgecolor", "black")

        if colour_by_composition:
            point_colours = tie_line[:, 0]
            cmap = conc_cmap
        else:
            point_colours = hull.colours[1]
            cmap = None
        ax.scatter(
            tie_line[:, 0],
            tie_line[:, 1],
            c=point_colours,
            cmap=cmap,
            marker="o",
            zorder=99999,
            edgecolor=edgecolor,
            s=scale * hull_point_scale,
            lw=edgewidth,
        )
        if plot_tie_line:
            ax.plot(
                np.sort(tie_line[:, 0]),
                tie_line[np.argsort(tie_line[:, 0]), 1],
                c=hull.colours[0],
                zorder=1,
                label=hull_label,
                marker="o",
                markerfacecolor=hull.colours[0],
                markeredgecolor=edgecolor,
                markeredgewidth=edgewidth,
                markersize=np.sqrt(scale * hull_point_scale),
            )
    if plot_tie_line:
        ax.plot(
            np.sort(tie_line[:, 0]),
            tie_line[np.argsort(tie_line[:, 0]), 1],
            c=hull.colours[0],
            zorder=1,
            label=hull_label,
            markersize=0,
        )

    if hull.hull_cutoff > 0:
        ax.plot(
            np.sort(tie_line[:, 0]),
            tie_line[np.argsort(tie_line[:, 0]), 1] + hull.hull_cutoff,
            "--",
            c=hull.colours[1],
            alpha=0.5,
            zorder=1,
            label="",
        )

    # annotate hull structures
    if labels or label_cutoff is not None:
        label_cursor = _get_hull_labels(hull, num_species=2, label_cutoff=label_cutoff)
        already_labelled = []
        for ind, doc in enumerate(label_cursor):
            formula = get_formula_from_stoich(doc["stoichiometry"], sort=True)
            if formula not in already_labelled:
                arrowprops = dict(
                    arrowstyle="-|>",
                    facecolor="k",
                    edgecolor="k",
                    lw=2,
                    alpha=1,
                    zorder=1,
                    shrinkA=2,
                    shrinkB=4,
                )
                min_comp = tie_line[np.argmin(tie_line[:, 1]), 0]
                e_f = label_cursor[ind]["formation_" + str(hull.energy_key)]
                conc = label_cursor[ind]["concentration"][0]
                if conc < min_comp:
                    position = (0.8 * conc, label_offset[0] * (e_f - label_offset[1]))
                elif label_cursor[ind]["concentration"][0] == min_comp:
                    position = (conc, label_offset[0] * (e_f - label_offset[1]))
                else:
                    position = (
                        min(1.1 * conc + 0.15, 0.95),
                        label_offset[0] * (e_f - label_offset[1]),
                    )
                    position = (
                        min(1.05 * conc + 0.15, 0.95),
                        label_offset[0] * (e_f - label_offset[1]),
                    )
                if _specific_label_offset:
                    plain_formula = get_formula_from_stoich(
                        doc["stoichiometry"], tex=False
                    )
                    if plain_formula in _specific_label_offset:
                        offset = _specific_label_offset.pop(plain_formula)
                        if offset is None:
                            continue
                        position = (position[0] + offset[0], position[1] + offset[1])
                if colour_by_composition:
                    text_colour = conc_cmap(conc)
                else:
                    text_colour = "k"
                ax.annotate(
                    get_formula_from_stoich(
                        doc["stoichiometry"],
                        latex_sub_style=r"\mathregular",
                        tex=True,
                        elements=hull.species,
                        sort=False,
                    ),
                    xy=(conc, e_f),
                    xytext=position,
                    color=text_colour,
                    textcoords="data",
                    ha="right",
                    va="bottom",
                    arrowprops=arrowprops,
                    zorder=1,
                )
                already_labelled.append(formula)

    if _specific_label_offset:
        import warnings

        warnings.warn(f"Found unused requested offsets: {_specific_label_offset}")

    # points for off hull structures; we either colour by source or by energy
    if plot_points and not colour_by_source:

        if hull.hull_cutoff == 0:
            # if no specified hull cutoff, ignore labels and colour by hull distance
            if plot_points:
                concs = hull.structures[np.argsort(hull.hull_dist), 0][::-1]
                energies = (hull.structures[np.argsort(hull.hull_dist), -1][::-1],)
                if colour_by_composition:
                    point_colours = concs
                    norm = None
                    cmap = conc_cmap
                else:
                    point_colours = np.sort(hull.hull_dist)[::-1]
                    norm = matplotlib.colors.LogNorm(0.01, 1, clip=True)
                    cmap = hull.default_cmap
                scatter = ax.scatter(
                    concs,
                    energies,
                    s=scale * 40,
                    c=point_colours,
                    zorder=100,
                    cmap=cmap,
                    norm=norm,
                    rasterized=True,
                )

                if show_cbar:
                    ticks = [0.01, 0.02, 0.04, 0.08, 0.16, 0.32, 0.64]
                    cbar = plt.colorbar(
                        scatter, aspect=30, pad=0.02, ticks=ticks, extend="both"
                    )
                    if hull_dist_unit.lower() == "mev":
                        ticks = [int(1000 * t) for t in ticks]
                    cbar.ax.set_yticklabels(ticks)
                    cbar.ax.tick_params(which="minor", length=0)
                    unit = "eV/atom"
                    if hull_dist_unit.lower() == "mev":
                        unit = "m" + unit
                    cbar.set_label(f"Distance from hull ({unit})")

        elif hull.hull_cutoff != 0:
            # if specified hull cutoff colour those below
            c = hull.colours[1]
            for ind in range(len(hull.structures)):
                if hull.hull_dist[ind] <= hull.hull_cutoff or hull.hull_cutoff == 0:
                    if plot_points:
                        scatter.append(
                            ax.scatter(
                                hull.structures[ind, 0],
                                hull.structures[ind, 1],
                                s=scale * 40,
                                alpha=0.9,
                                c=c,
                                zorder=300,
                            )
                        )
            if plot_points:
                ax.scatter(
                    hull.structures[1:-1, 0],
                    hull.structures[1:-1, 1],
                    s=scale * 30,
                    lw=0,
                    alpha=0.3,
                    c=hull.colours[-2],
                    edgecolor="k",
                    zorder=10,
                    rasterized=True,
                )

    elif colour_by_source:
        _scatter_plot_by_source(
            hull,
            ax,
            scale,
            kwargs,
            sources=sources,
            source_labels=source_labels,
            plot_hull_points=plot_hull_points,
            legend_kwargs=legend_kwargs,
        )

    if eform_limits is None:
        eform_limits = (np.min(hull.structures[:, 1]), np.max(hull.structures[:, 1]))
        lims = (
            -0.1 if eform_limits[0] >= 0 else 1.25 * eform_limits[0],
            eform_limits[1] if eform_limits[0] >= 0 else 0.05,
        )
    else:
        lims = sorted(eform_limits)
    ax.set_ylim(lims)

    if isinstance(title, bool) and title:
        if hull._non_elemental:
            ax.set_title(
                r"({d[0]})$_\mathrm{{x}}$({d[1]})$_\mathrm{{1-x}}$".format(
                    d=chempot_labels
                )
            )
        else:
            ax.set_title(
                r"{d[0]}$_\mathrm{{x}}${d[1]}$_\mathrm{{1-x}}$".format(d=chempot_labels)
            )
    elif isinstance(title, str) and title != "":
        ax.set_title(title)

    if hull._non_elemental:
        ax.set_xlabel(
            r"x in ({d[0]})$_\mathrm{{x}}$({d[1]})$_\mathrm{{1-x}}$".format(
                d=chempot_labels
            )
        )
    else:
        ax.set_xlabel(
            r"x in {d[0]}$_\mathrm{{x}}${d[1]}$_\mathrm{{1-x}}$".format(
                d=chempot_labels
            )
        )

    ax.set_xlim(-0.05, 1.05)
    ax.set_ylabel("Formation energy (eV/atom)")

    if hull.savefig or any([kwargs.get(ext) for ext in SAVE_EXTS]):
        import os

        fname = plot_fname or "".join(hull.species) + "_hull"
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                fname_tmp = fname
                ind = 0
                while os.path.isfile("{}.{}".format(fname_tmp, ext)):
                    ind += 1
                    fname_tmp = fname + str(ind)

                fname = fname_tmp
                plt.savefig(
                    "{}.{}".format(fname, ext), bbox_inches="tight", transparent=True
                )
                print("Wrote {}.{}".format(fname, ext))

    return ax


@plotting_function
def plot_ensemble_hull(
    hull,
    data_key,
    ax=None,
    formation_energy_key="formation_enthalpy_per_atom",
    plot_points=False,
    plot_hull_points=True,
    alpha_scale=0.25,
    plot_hulls=True,
    **kwargs,
):
    """Plot and generate an ensemble of hulls. If axis not requested,
    a histogram of frequency of a particular concentration appearing on
    the convex hull is also generated as a second axis.

    Parameters:
        hull (QueryConvexHull): hull object created with a cursor that
            contains the data key for all entries in hull.cursor.
        data_key (str): the key under which ensemble data is stored.

    Keyword arguments:
        ax (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        formation_energy_key (str): the key under which formation energies have been stored.
        alpha_scale (float): value by which to scale transparency of hulls.
        plot_points (bool): whether to plot the hull points for each hull in the ensemble.
        plot_hulls (bool): whether to plot the hull tie-lines for each hull in the ensemble.
        voltages (bool): compute average voltage and heatmaps.

    """
    import matplotlib.pyplot as plt

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    n_hulls = len(hull.phase_diagrams)
    plot_2d_hull(
        hull, ax=ax, plot_points=False, plot_hull_points=True, show=False, **kwargs
    )
    min_ef = 0
    colours_list = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])
    alpha = min([1, max([1 / (alpha_scale * n_hulls), 0.01])])
    for ind, _ in enumerate(hull.phase_diagrams):
        hull_cursor = [
            doc
            for doc in hull.cursor
            if doc[data_key]["hull_distance"][ind] <= 0.0 + EPS
        ]
        min_ef = np.min(
            [doc[data_key][formation_energy_key][ind] for doc in hull_cursor] + [min_ef]
        )
        if plot_hulls:
            ax.plot(
                [doc["concentration"][0] for doc in hull_cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                alpha=alpha,
                c="k",
                lw=0.5,
                zorder=0,
            )
        if plot_hull_points:
            ax.scatter(
                [doc["concentration"][0] for doc in hull_cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                alpha=alpha,
                marker="o",
                c=colours_list[1],
                lw=0,
                zorder=0,
            )
        if plot_points:
            ax.scatter(
                [doc["concentration"][0] for doc in hull.cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull.cursor],
                alpha=alpha,
                marker="o",
                c="k",
                s=5,
                lw=0,
                zorder=0,
            )

    ax.set_ylim(1.1 * min_ef)


@plotting_function
def plot_temperature_hull(
    hull,
    cmap="plasma",
    cmap_limits=(0.2, 0.8),
    ax=None,
    formation_energy_key="formation_free_energy_per_atom",
    plot_points=False,
    show_cbar=True,
    plot_hull_points=True,
    alpha_scale=1,
    lw_scale=1,
    plot_hulls=True,
    plot_fname=None,
    **kwargs,
):
    """Plot and generate an ensemble of hulls. If axis not requested,
    a histogram of frequency of a particular concentration appearing on
    the convex hull is also generated as a second axis.

    Parameters:
        hull (QueryConvexHull): hull object created with a cursor that
            contains the data key for all entries in hull.cursor.

    Keyword arguments:
        ax (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        formation_energy_key (str): the key under which formation energies have been stored.
        alpha_scale (float): value by which to scale transparency of hulls.
        plot_points (bool): whether to plot the hull points for each hull in the ensemble.
        plot_hulls (bool): whether to plot the hull tie-lines for each hull in the ensemble.
        voltages (bool): compute average voltage and heatmaps.

    """
    import matplotlib.pyplot as plt
    from matplotlib.colors import rgb2hex

    data_key = hull.data_key

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    n_hulls = len(hull.phase_diagrams)
    colours = [
        rgb2hex(col)
        for col in plt.cm.get_cmap(cmap)(np.linspace(*cmap_limits, n_hulls)).tolist()
    ]

    min_ef = 0
    alpha = alpha_scale

    # hack the energy key so that labels work
    _cached_key = hull.energy_key
    hull.energy_key = hull.chempot_energy_key
    max_temperature = max(hull.temperatures)
    # set up initial plot without plotting tie line
    ax = plot_2d_hull(
        hull,
        ax=ax,
        plot_tie_line=False,
        plot_points=False,
        plot_hull_points=False,
        show=False,
        **kwargs,
    )
    hull.energy_key = _cached_key

    ax.plot(
        [doc["concentration"][0] for doc in hull.hull_cursor],
        [doc["formation_" + hull.chempot_energy_key] for doc in hull.hull_cursor],
        marker="o",
        alpha=1,
        c="k",
        lw=2 * lw_scale,
        ls="--",
        label="Static",
        zorder=1e5,
    )

    ind = 0
    hull_cursor = [
        doc for doc in hull.cursor if doc[data_key]["hull_distance"][ind] <= 0.0 + EPS
    ]
    ax.plot(
        [doc["concentration"][0] for doc in hull_cursor],
        [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
        marker="o",
        c=colours[ind],
        lw=2 * lw_scale,
        markeredgewidth=1.5,
        markeredgecolor="k",
        ls="--",
        zorder=1e5,
        label="Static + ZPE",
    )

    # plot remaining temperatures
    for ind, _ in enumerate(hull.temperatures[1:]):
        hull_cursor = [
            doc
            for doc in hull.cursor
            if doc[data_key]["hull_distance"][ind] <= 0.0 + EPS
        ]
        min_ef = np.min(
            [doc[data_key][formation_energy_key][ind] for doc in hull_cursor] + [min_ef]
        )
        colour_ind = int(n_hulls * hull.temperatures[ind] / max_temperature)
        if plot_hulls:
            ax.plot(
                [doc["concentration"][0] for doc in hull_cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                alpha=alpha,
                c=colours[colour_ind],
                lw=1 * lw_scale,
                zorder=0,
            )
        if plot_hull_points:
            ax.scatter(
                [doc["concentration"][0] for doc in hull_cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull_cursor],
                alpha=1,
                marker="o",
                c=colours[colour_ind],
                lw=0.5,
                edgecolor="k",
                label="Structures on hull" if ind == n_hulls - 2 else None,
                zorder=1e4,
            )
        if plot_points:
            ax.scatter(
                [doc["concentration"][0] for doc in hull.cursor],
                [doc[data_key][formation_energy_key][ind] for doc in hull.cursor],
                label="Structures above hull" if ind == n_hulls - 2 else None,
                alpha=1,
                marker="o",
                edgecolor="w",
                c=colours[colour_ind],
                lw=0.5,
                zorder=1e-3,
            )

    ax.set_ylim(1.1 * min_ef)

    if show_cbar:
        import matplotlib.colors

        mappable = plt.cm.ScalarMappable(
            cmap=matplotlib.colors.LinearSegmentedColormap.from_list("cut", colours),
            norm=plt.Normalize(vmin=0, vmax=np.max(hull.temperatures)),
        )
        mappable._A = hull.temperatures
        cbar = plt.colorbar(mappable, alpha=alpha)
        cbar.ax.tick_params(length=0)
        cbar.ax.yaxis.set_ticks_position("right")
        cbar.ax.set_frame_on(False)
        cbar.outline.set_visible(False)
        cbar.set_label("Temperature (K)")

    if hull.savefig or any(kwargs.get(ext) for ext in SAVE_EXTS):
        if plot_fname is not None:
            fname = plot_fname
        else:
            fname = "".join(hull.species) + data_key + "_hull"
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig(
                    "{}.{}".format(fname, ext), bbox_inches="tight", transparent=True
                )
                print("Wrote {}.{}".format(fname, ext))

    ax.legend()

    return ax


@plotting_function
def plot_ternary_hull(
    hull,
    axis=None,
    show=True,
    plot_points=True,
    hull_cutoff=None,
    fig_height=None,
    fig_width=None,
    label_cutoff=None,
    label_corners=True,
    expecting_cbar=True,
    labels=None,
    plot_fname=None,
    colour_points_by="hull_distance",
    hull_dist_unit="meV",
    efmap=None,
    sampmap=None,
    concmap=None,
    capmap=None,
    pathways=False,
    scale_override=None,
    label_spacing=0.05,
    label_offset=0.8,
    **kwargs,
):
    """Plot calculated ternary hull as a 2D projection.

    Parameters:
        hull (matador.hull.QueryConvexHull): matador hull object.

    Keyword arguments:
        axis (matplotlib.axes.Axes): matplotlib axis object on which to plot.
        show (bool): whether or not to show plot in X window.
        plot_points (bool): whether or not to plot each structure as a point.
        colour_points_by (str): either `hull_distance` or `concentration`.
        label_cutoff (float/:obj:`tuple` of :obj:`float`): draw labels less than or
            between these distances form the hull, also read from hull.args.
        expecting_cbar (bool): whether or not to space out the plot to preserve
            aspect ratio if a colourbar is present.
        labels (bool): whether or not to label on-hull structures
        label_corners (bool): whether or not to put axis labels on corners or edges.
        label_offset (float): Location of the top left rcorner of the labels.
        label_spacing (float): Spacing between label list.
        hull_dist_unit (str): either "eV" or "meV",
        png/pdf/svg (bool): whether or not to write the plot to a file.
        plot_fname (str): filename to write plot to.
        efmap (bool/str): plot heatmap of formation energy,
        sampmap (bool/str): plot heatmap showing sampling density,
        concmap (bool/str): colour background by concentration and VESTA colours
        capmap (bool/str): plot heatmap showing gravimetric capacity.
        pathways (bool): plot the pathway from the starting electrode to active ion.

    Returns:
        matplotlib.axes.Axes: matplotlib axis with plot.

    """
    import ternary
    import matplotlib.pyplot as plt
    import matplotlib.colors
    from matador.utils.chem_utils import get_generic_grav_capacity

    _colour_points_by_values = ("hull_distance", "concentration")
    if colour_points_by not in _colour_points_by_values:
        raise RuntimeError(
            f"colour_points_by` must be in {_colour_points_by_values}, not {colour_points_by}"
        )

    if efmap is None:
        efmap = hull.args.get("efmap")
    if sampmap is None:
        sampmap = hull.args.get("sampmap")
    if capmap is None:
        capmap = hull.args.get("capmap")
    if pathways is None:
        pathways = hull.args.get("pathways")

    if labels is None:
        labels = hull.args.get("labels")
    if label_cutoff is None:
        label_cutoff = hull.args.get("label_cutoff")
        if label_cutoff is None:
            label_cutoff = 0
    else:
        labels = True

    _hull_cutoff = 0.1
    if hull_cutoff is not None:
        _hull_cutoff = hull_cutoff
    elif hull.hull_cutoff is not None:
        _hull_cutoff = hull.hull_cutoff
    hull_cutoff = _hull_cutoff

    print("Plotting ternary hull...")
    if capmap or efmap or concmap:
        scale = 100
    elif sampmap:
        scale = 25
    else:
        scale = 1

    if scale_override:
        scale = scale_override

    if axis is not None:
        fig, ax = ternary.figure(scale=scale, ax=axis)
    else:
        fig, ax = ternary.figure(scale=scale)

    # maintain aspect ratio of triangle
    if fig_height is None:
        _user_height = plt.rcParams.get("figure.figsize", (8, 6))[0]
    else:
        _user_height = fig_height
    if fig_width is None:
        if capmap or efmap or sampmap:
            _user_width = 8 / 5 * _user_height
        elif not expecting_cbar:
            _user_width = _user_height
        else:
            _user_width = 6.67 / 5 * _user_height
    else:
        _user_width = fig_width

    fig.set_size_inches(_user_width, _user_height)

    ax.boundary(linewidth=2.0, zorder=99)
    ax.get_axes().axis("off")
    ax.clear_matplotlib_ticks()

    chempot_labels = [
        get_formula_from_stoich(
            get_stoich_from_formula(species, sort=False), sort=False, tex=True
        )
        for species in hull.species
    ]

    ax.gridlines(color="black", multiple=scale * 0.1, linewidth=0.5)
    ticks = [float(val) for val in np.linspace(0, 1, 6)]

    element_colours = [get_element_colours()[s] for s in hull.species]
    if label_corners:
        # remove 0 and 1 ticks when labelling corners
        ticks = ticks[1:-1]
        ax.left_corner_label(
            chempot_labels[2], fontsize="x-large", color=element_colours[2]
        )
        ax.right_corner_label(
            chempot_labels[0], fontsize="x-large", color=element_colours[0]
        )
        ax.top_corner_label(
            chempot_labels[1], fontsize="x-large", color=element_colours[1], offset=0.16
        )
    else:
        ax.left_axis_label(chempot_labels[2], fontsize="large", offset=0.12)
        ax.right_axis_label(chempot_labels[1], fontsize="large", offset=0.12)
        ax.bottom_axis_label(chempot_labels[0], fontsize="large", offset=0.08)
        ax.set_title(
            "-".join(["{}".format(label) for label in chempot_labels]),
            fontsize="large",
            y=1.02,
        )

    ax.ticks(
        axis="lbr",
        linewidth=1,
        offset=0.025,
        fontsize="medium",
        locations=(scale * np.asarray(ticks)).tolist(),
        ticks=ticks,
        tick_formats="%.1f",
    )

    concs = np.zeros((len(hull.structures), 3))
    concs[:, :-1] = hull.structures[:, :-1]
    for i in range(len(concs)):
        # set third triangular coordinate
        concs[i, -1] = 1 - concs[i, 0] - concs[i, 1]

    stable = concs[np.where(hull.hull_dist <= 0 + EPS)]

    # sort by hull distances so things are plotting the right order
    concs = concs[np.argsort(hull.hull_dist)].tolist()
    hull_dist = np.sort(hull.hull_dist)

    filtered_concs = []
    filtered_hull_dists = []
    for ind, conc in enumerate(concs):
        if conc not in filtered_concs:
            if hull_dist[ind] <= hull_cutoff:
                filtered_concs.append(conc)
                filtered_hull_dists.append(hull_dist[ind])

    unfiltered_concs = None
    if sampmap:
        unfiltered_concs = np.asarray(concs)

    concs = np.asarray(filtered_concs)
    hull_dist = np.asarray(filtered_hull_dists)

    min_cut = 0.0
    max_cut = 0.1

    if hull_dist_unit.lower() == "mev":
        hull_dist *= 1000
        min_cut *= 1000
        max_cut *= 1000

    if not hasattr(hull, "colours"):
        hull.colours = list(plt.rcParams["axes.prop_cycle"].by_key()["color"])
    hull.default_cmap_list = get_linear_cmap(hull.colours[1:4], list_only=True)
    hull.default_cmap = get_linear_cmap(hull.colours[1:4], list_only=False)
    n_colours = len(hull.default_cmap_list)
    colours_hull = hull.default_cmap_list
    cmap = hull.default_cmap

    for plane in hull.convex_hull.planes:
        plane.append(plane[0])
        plane = np.asarray(plane)
        ax.plot(
            scale * plane,
            c=kwargs.get("tielinecolor", hull.colours[0]),
            lw=1.5,
            alpha=1,
            zorder=98,
        )

    if pathways:
        for phase in stable:
            if phase[0] == 0 and phase[1] != 0 and phase[2] != 0:
                ax.plot(
                    [scale * phase, [scale, 0, 0]], c="r", alpha=0.2, lw=6, zorder=99
                )

    # add points
    if plot_points:
        colours_list = []
        colour_metric = hull_dist
        conc_colours = []
        for i, _ in enumerate(colour_metric):
            if colour_points_by == "concentration":
                conc_colours.append(
                    colour_from_ternary_concentration(concs[i], hull.species)
                )
            if hull_dist[i] >= max_cut:
                colours_list.append(n_colours - 1)
            elif hull_dist[i] <= min_cut:
                colours_list.append(0)
            else:
                colours_list.append(int((n_colours - 1) * (hull_dist[i] / max_cut)))
        colours_list = np.asarray(colours_list)
        if colour_points_by == "concentration":
            colours = [
                colour_from_ternary_concentration(stable[i], hull.species)
                for i in range(len(stable))
            ]
        else:
            colours = hull.colours[1]

        # Plot stable points
        ax.scatter(
            scale * stable,
            marker="o",
            color=colours,
            edgecolors=kwargs.get("markeredgecolor", "black"),
            zorder=9999999,
            s=150,
            lw=1.5,
        )

        # Plot unstable points
        for i, _ in enumerate(concs):
            if colour_points_by == "concentration":
                c = conc_colours[i]
            else:
                c = colours_hull[colours_list[i]]

            ax.scatter(
                scale * concs[i].reshape(1, 3),
                color=c,
                marker="o",
                zorder=10000 - colours_list[i],
                s=70 * (1 - float(colours_list[i]) / n_colours) + 15,
                lw=1,
                edgecolors=kwargs.get("markeredgecolor", "black"),
            )

        # Test points for colourbar normalization
        if colour_points_by == "hull_distance":
            ax.scatter(
                scale * concs,
                colormap=cmap,
                colorbar=(colour_points_by == "hull_distance"),
                cbarlabel="Distance from hull ({}eV/atom)".format(
                    "m" if hull_dist_unit.lower() == "mev" else ""
                ),
                c=colour_metric,
                vmax=max_cut,
                vmin=min_cut,
                zorder=1000,
                s=40,
                alpha=0,
            )

        # Set up legend in case of colour by concentration
        elif colour_points_by == "concentration":
            labelled_hull_dists = [0, 25, 100]
            for dist in labelled_hull_dists:
                if dist != 0:
                    size = (
                        70 * (1 - int((n_colours - 1) * (dist / max_cut)) / n_colours)
                        + 15
                    )
                    lw = 1
                else:
                    size = 150
                    lw = 1.5
                ax.scatter(
                    [[scale, 0, 0]],
                    color="lightgrey",
                    marker="o",
                    zorder=0,
                    s=size,
                    lw=lw,
                    label=f"{dist:.0f} meV/atom",
                    edgecolors="grey",
                )

            ax.legend(title="Hull distance", fontsize="medium")

    # Add colourmaps for backgrounds
    if capmap:
        capacities = dict()
        from ternary.helpers import simplex_iterator

        for (i, j, k) in simplex_iterator(scale):
            capacities[(i, j, k)] = get_generic_grav_capacity(
                [float(i) / scale, float(j) / scale, float(scale - i - j) / scale],
                hull.species,
            )

        if not isinstance(capmap, str):
            capmap = "Pastel2"

        cmap_full = plt.cm.get_cmap(capmap)
        if cmap_full.N < 255:
            cmap_full = matplotlib.colors.LinearSegmentedColormap.from_list(
                capmap, cmap_full.colors
            )

        ax.heatmap(
            capacities,
            style="hexagonal",
            cbarlabel="Gravimetric capacity (mAh/g)",
            vmin=0,
            vmax=3000,
            cmap=cmap_full,
        )
    elif efmap:
        energies = dict()
        fake_structures = []
        from ternary.helpers import simplex_iterator

        for (i, j, k) in simplex_iterator(scale):
            fake_structures.append([float(i) / scale, float(j) / scale, 0.0])
        fake_structures = np.asarray(fake_structures)
        plane_energies = hull.get_hull_distances(fake_structures, precompute=False)
        ind = 0
        for (i, j, k) in simplex_iterator(scale):
            energies[(i, j, k)] = -1 * plane_energies[ind]
            ind += 1

        if not isinstance(efmap, str):
            efmap = "BuPu_r"

        ax.heatmap(
            energies,
            style="hexagonal",
            cbarlabel="Formation energy (eV/atom)",
            vmax=0,
            cmap=efmap,
        )
    elif sampmap:
        if not isinstance(sampmap, str):
            sampmap = "plasma"

        sampling = dict()
        from ternary.helpers import simplex_iterator

        eps = 1.0 / float(scale)
        for (i, j, k) in simplex_iterator(scale):
            sampling[(i, j, k)] = np.size(
                np.where(
                    (unfiltered_concs[:, 0] <= float(i) / scale + eps)
                    * (unfiltered_concs[:, 0] >= float(i) / scale - eps)
                    * (unfiltered_concs[:, 1] <= float(j) / scale + eps)
                    * (unfiltered_concs[:, 1] >= float(j) / scale - eps)
                    * (unfiltered_concs[:, 2] <= float(k) / scale + eps)
                    * (unfiltered_concs[:, 2] >= float(k) / scale - eps)
                )
            )
        ax.heatmap(
            sampling,
            style="hexagonal",
            cbarlabel="Number of structures",
            cmap=sampmap,
        )
    elif concmap:
        concs = dict()
        from ternary.helpers import simplex_iterator

        for (i, j, k) in simplex_iterator(scale):
            colour = colour_from_ternary_concentration(
                [float(i) / scale, float(j) / scale, float(k) / scale],
                hull.species,
                alpha=0.5,
            )
            concs[(i, j, k)] = colour
        ax.heatmap(concs, style="hexagonal", use_rgba=True, colorbar=False)

    # add labels
    if labels:
        label_cursor = _get_hull_labels(hull, label_cutoff=label_cutoff, num_species=3)
        label_coords = [
            [0.0, label_offset - i * label_spacing] for i in range(len(label_cursor))
        ]

        for ind, doc in enumerate(label_cursor):
            conc = np.asarray(doc["concentration"] + [1 - sum(doc["concentration"])])
            label = get_formula_from_stoich(
                doc["stoichiometry"],
                sort=False,
                tex=True,
                latex_sub_style=r"\mathregular",
                elements=hull.species,
            )
            if label_cutoff > 0 and doc["hull_distance"] > EPS:
                label += f' ({1000 * doc["hull_distance"]:.1f} meV/atom)'
            arrowprops = dict(
                arrowstyle="-|>", lw=2, alpha=0, zorder=1, shrinkA=2, shrinkB=4
            )
            if colour_points_by == "concentration":
                text_colour = colour_from_ternary_concentration(conc, hull.species)
            else:
                text_colour = "k"

            ax.annotate(
                label,
                scale * conc,
                textcoords="data",
                xytext=[scale * val for val in label_coords[ind]],
                ha="left",
                va="bottom",
                color=text_colour,
                arrowprops=arrowprops,
            )

    plt.tight_layout(w_pad=0.2)
    # important for retaining labels if exporting to PDF
    # see https://github.com/marcharper/python-ternary/issues/36
    ax.ax.set_aspect("equal", adjustable="box")
    ax._redraw_labels()  # noqa

    if hull.savefig or any(kwargs.get(ext) for ext in SAVE_EXTS):
        fname = plot_fname or "".join(hull.species) + "_hull"
        for ext in SAVE_EXTS:
            if hull.args.get(ext) or kwargs.get(ext):
                plt.savefig(
                    "{}.{}".format(fname, ext), bbox_inches="tight", transparent=True
                )
                print("Wrote {}.{}".format(fname, ext))
    elif show:
        print("Showing plot...")
        plt.show()

    return ax


def _scatter_plot_by_source(
    hull,
    ax,
    scale,
    kwargs,
    sources=None,
    source_labels=None,
    plot_hull_points=True,
    legend_kwargs=None,
):
    """Add scatter points to the hull depending on the guessed
    provenance of a structure.

    """
    from matador.utils.cursor_utils import get_guess_doc_provenance

    if sources is None:
        sources = [
            "AIRSS",
            "GA",
            "OQMD",
            "SWAPS",
            "ICSD",
            "DOI",
            "SM",
            "MP",
            "PF",
            "Other",
        ]

    if source_labels is None:
        source_labels = sources
    else:
        assert len(source_labels) == len(sources)

    if "Other" not in sources:
        sources.append("Other")
        source_labels.append("Other")

    # hack: double length of hull colours
    hull.colours.extend(hull.colours)

    colour_choices = {
        source: hull.colours[ind + 1] for ind, source in enumerate(sources)
    }
    points_by_source = {source: defaultdict(list) for source in sources}
    hull_points_by_source = {source: defaultdict(list) for source in sources}
    sources_present = set()
    for doc in hull.cursor:
        source = get_guess_doc_provenance(doc["source"])
        if source not in sources:
            # use grey for undesired sources
            source = "Other"
        if doc["hull_distance"] <= 0 + 2e-3 and not plot_hull_points:
            hull_points_by_source[source]["concs"].append(doc["concentration"])
            hull_points_by_source[source]["energies"].append(
                doc["formation_{}".format(hull.energy_key)]
            )
        else:
            points_by_source[source]["concs"].append(doc["concentration"])
            points_by_source[source]["energies"].append(
                doc["formation_{}".format(hull.energy_key)]
            )

        sources_present.add(source)

    alpha = kwargs.get("alpha")
    if alpha is None:
        alpha = 0.2

    legend_sources = {}

    for source in sources:
        if "concs" not in points_by_source[source]:
            continue

        concs = points_by_source[source]["concs"]
        energies = points_by_source[source]["energies"]
        ax.scatter(
            concs,
            energies,
            c=colour_choices[source],
            alpha=alpha,
            s=scale * 20,
            lw=0,
            zorder=100,
            rasterized=True,
        )

        legend_sources[source] = colour_choices[source]

    hull_point_options = dict(edgecolor="k", alpha=1, s=scale * 40, lw=1.5, zorder=1e5)

    if not plot_hull_points:
        for source in sources:
            if "concs" not in hull_points_by_source[source]:
                continue

            concs = hull_points_by_source[source]["concs"]
            energies = hull_points_by_source[source]["energies"]
            ax.scatter(
                concs, energies, facecolor=colour_choices[source], **hull_point_options
            )

            legend_sources[source] = colour_choices[source]

    for ind, source in enumerate(sources):
        if source in legend_sources:
            ax.scatter(
                1e10,
                1e10,
                facecolor=legend_sources[source],
                label=source_labels[ind],
                **hull_point_options,
            )

    if legend_kwargs is not None:
        legend = ax.legend(**legend_kwargs)
    else:
        legend = ax.legend(ncol=2)
    legend.set_zorder(1e20)

    return ax



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the CastepSpectralWorkflow, which performs
spectral calculations with CASTEP in multiple steps:

1. Performs a singlepoint calculation (if check file is not found).
2. If the ``spectral_kpoints_mp_spacing`` keyword is found, interpolate
   wavefunction to DOS grid.
   - If an OptaDOS input file (.odi) with the root seedname
     is found, run OptaDOS on the resulting density of states.
3. If ``spectral_kpoints_path_spacing`` keyword is present, create
   a bandstructure on the SeeK-path generated path.

"""


import os
import copy
import glob
import logging
from matador.workflows import Workflow
from matador.scrapers import arbitrary2dict
from matador.workflows.castep.common import castep_scf

LOG = logging.getLogger("run3")


def castep_full_spectral(computer, calc_doc, seed, **kwargs):
    """Perform a "full" spectral calculation on a system, i.e. first
    perform an SCF then interpolate to different kpoint paths/grids to
    form DOS and dispersions. Optionally use OptaDOS for post-processing
    of DOS.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepSpectralWorkflow(computer, calc_doc, seed)
    return workflow.success


class CastepSpectralWorkflow(Workflow):
    """Perform a "full" spectral calculation on a system, i.e. first
    perform an SCF then interpolate to different kpoint paths/grids to
    form DOS and dispersions. Optionally use OptaDOS for post-processing
    of DOS.

    Attributes:
        computer (:obj:`matador.compute.ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.

    """

    def preprocess(self):
        """Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """
        # default todo
        todo = {
            "scf": True,
            "dos": False,
            "pdos": False,
            "broadening": False,
            "dispersion": False,
            "pdis": False,
        }
        # definition of steps and names
        steps = {
            "scf": castep_spectral_scf,
            "dos": castep_spectral_dos,
            "pdos": optados_pdos,
            "broadening": optados_dos_broadening,
            "dispersion": castep_spectral_dispersion,
            "pdis": optados_pdispersion,
        }

        exts = {
            "scf": {
                "input": [".cell", ".param"],
                "output": [".castep", ".*err", "-out.cell"],
            },
            "dos": {
                "input": [".cell", ".param"],
                "output": [
                    ".castep",
                    ".bands",
                    ".pdos_bin",
                    ".dome_bin",
                    ".*err",
                    "-out.cell",
                ],
            },
            "dispersion": {
                "input": [".cell", ".param"],
                "output": [
                    ".castep",
                    ".bands",
                    ".pdos_bin",
                    ".dome_bin",
                    ".*err",
                    "-out.cell",
                ],
            },
            "pdis": {"input": [".odi", ".pdos_bin"], "output": [".odo", ".*err"]},
            "pdos": {
                "input": [".odi", ".pdos_bin", ".dome_bin"],
                "output": [".odo", ".*err"],
            },
            "broadening": {
                "input": [".odi", ".pdos_bin", ".dome_bin"],
                "output": [".odo", ".*err"],
            },
        }

        if os.path.isfile(self.seed + ".check"):
            LOG.info("Found {}.check, so skipping initial SCF.".format(self.seed))
            todo["scf"] = False

        if (
            "spectral_kpoints_path" in self.calc_doc
            or "spectral_kpoints_list" in self.calc_doc
            or "spectral_kpoints_path_spacing" in self.calc_doc
            or self.calc_doc.get("spectral_task", "").lower() == "bandstructure"
        ):
            todo["dispersion"] = not os.path.isfile(self.seed + ".bands_dispersion")

        if (
            "spectral_kpoints_mp_spacing" in self.calc_doc
            or self.calc_doc.get("spectral_task", "").lower() == "dos"
        ):
            todo["dos"] = not os.path.isfile(self.seed + ".bands_dos")

        odi_fname = _get_optados_fname(self.seed)
        if odi_fname is not None:
            odi_dict, _ = arbitrary2dict(odi_fname)
            if todo["dispersion"]:
                todo["pdis"] = "pdispersion" in odi_dict
            if todo["dos"]:
                todo["broadening"] = "broadening" in odi_dict
                todo["pdos"] = "pdos" in odi_dict

        for key in todo:
            if todo[key]:
                self.add_step(
                    steps[key],
                    key,
                    input_exts=exts[key].get("input"),
                    output_exts=exts[key].get("output"),
                )

        # if not using a user-requested path, use seekpath and spglib
        # to reduce to primitive and use consistent path
        if (
            "spectral_kpoints_list" not in self.calc_doc
            and "spectral_kpoints_path" not in self.calc_doc
        ):
            from matador.utils.cell_utils import cart2abc

            prim_doc, kpt_path = self.computer.get_seekpath_compliant_input(
                self.calc_doc, self.calc_doc.get("spectral_kpoints_path_spacing", 0.05)
            )
            self.calc_doc.update(prim_doc)
            self.calc_doc["lattice_abc"] = cart2abc(self.calc_doc["lattice_cart"])
            if todo["dispersion"]:
                self.calc_doc["spectral_kpoints_list"] = kpt_path
        elif todo["dispersion"] and "spectral_kpoints_path" in self.calc_doc:
            self._user_defined_kpt_path = True
            LOG.warning("Using user-defined k-point path for all structures.")
            self.calc_doc["spectral_kpoints_path_spacing"] = self.calc_doc.get(
                "spectral_kpoints_path_spacing", 0.05
            )

        if todo["dos"]:
            self.calc_doc["spectral_kpoints_mp_spacing"] = self.calc_doc.get(
                "spectral_kpoints_mp_spacing", 0.05
            )

        # always use continuation
        self.calc_doc["continuation"] = "default"


def castep_spectral_scf(computer, calc_doc, seed):
    """Run a singleshot SCF calculation.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP spectral SCF...")

    scf_doc = copy.deepcopy(calc_doc)
    scf_doc["write_checkpoint"] = "ALL"
    scf_doc["task"] = "singlepoint"
    scf_doc["write_cell_structure"] = True
    if "spectral_task" in scf_doc:
        del scf_doc["spectral_task"]

    required = []
    forbidden = [
        "spectral_task",
        "spectral_kpoints_list",
        "spectral_kpoints_path",
        "spectral_kpoints_mp_spacing",
        "spectral_kpoints_path_spacing",
    ]

    return castep_scf(
        computer, scf_doc, seed, required_keys=required, forbidden_keys=forbidden
    )


def castep_spectral_dos(computer, calc_doc, seed):
    """Runs a DOS interpolation on top of a completed SCF. If a single
    .odi file is found, run OptaDOS on the resulting DOS.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP spectral DOS calculation...")

    dos_doc = copy.deepcopy(calc_doc)
    dos_doc["task"] = "spectral"
    dos_doc["spectral_task"] = "dos"
    # disable checkpointing for BS/DOS by default, leaving just SCF
    dos_doc["write_checkpoint"] = "none"
    dos_doc["write_cell_structure"] = True
    dos_doc["pdos_calculate_weights"] = True

    required = ["spectral_kpoints_mp_spacing"]
    forbidden = [
        "spectral_kpoints_list",
        "spectral_kpoints_path",
        "spectral_kpoints_path_spacing",
    ]

    computer.validate_calc_doc(dos_doc, required, forbidden)
    success = computer.run_castep_singleshot(
        dos_doc, seed, keep=True, intermediate=True
    )

    return success


def castep_spectral_dispersion(computer, calc_doc, seed):
    """Runs a dispersion interpolation on top of a completed SCF calculation,
    optionally running orbitals2bands and OptaDOS projected dispersion.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP spectral dispersion calculation...")
    disp_doc = copy.deepcopy(calc_doc)
    disp_doc["task"] = "spectral"
    disp_doc["spectral_task"] = "bandstructure"

    # disable checkpointing for BS/DOS by default, leaving just SCF
    disp_doc["write_checkpoint"] = "none"
    disp_doc["pdos_calculate_weights"] = True
    disp_doc["write_cell_structure"] = True
    disp_doc["continuation"] = "default"

    required = []
    forbidden = ["spectral_kpoints_mp_spacing"]

    computer.validate_calc_doc(disp_doc, required, forbidden)
    success = computer.run_castep_singleshot(
        disp_doc, seed, keep=True, intermediate=True
    )

    if disp_doc.get("write_orbitals"):
        LOG.info("Planning to call orbitals2bands...")

        _cache_executable = copy.deepcopy(computer.executable)
        _cache_core = copy.deepcopy(computer.ncores)
        computer.ncores = 1
        computer.executable = "orbitals2bands"
        try:
            success = computer.run_generic(intermediate=True, mv_bad_on_failure=False)
        except Exception as exc:
            computer.executable = _cache_executable
            computer.ncores = _cache_core
            LOG.warning("Failed to call orbitals2bands, with error: {}".format(exc))

        computer.ncores = _cache_core
        computer.executable = _cache_executable

    return success


def optados_pdos(computer, _, seed):
    """Run an OptaDOS projected-DOS.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """

    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)

        odi_dict["task"] = "pdos"
        if "pdispersion" in odi_dict:
            del odi_dict["pdispersion"]

        LOG.info(
            "Performing OptaDOS pDOS calculation with parameters from {}".format(
                odi_fname
            )
        )
        success = _run_optados(computer, odi_dict, seed, suffix="dos")
        return success

    return None


def optados_dos_broadening(computer, _, seed):
    """Run an OptaDOS total DOS broadening.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """

    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)
        # if broadening keyword is present, try to run a normal DOS
        odi_dict["task"] = "dos"
        if "pdos" in odi_dict:
            del odi_dict["pdos"]
        if "pdispersion" in odi_dict:
            del odi_dict["pdispersion"]

        LOG.info(
            "Performing OptaDOS DOS broadening with parameters from {}".format(
                odi_fname
            )
        )
        return _run_optados(computer, odi_dict, seed, suffix="dos")

    return None


def optados_pdispersion(computer, _, seed):
    """Runs an OptaDOS projected dispersion calculation.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        _ : second parameter is required but ignored.
        seed (str): root filename of structure.

    """
    odi_fname = _get_optados_fname(seed)
    if odi_fname is not None:
        odi_dict, _ = arbitrary2dict(odi_fname)
        # if pdispersion keyword is present, try to run a pdis
        odi_dict["task"] = "pdispersion"
        if "pdos" in odi_dict:
            del odi_dict["pdos"]

        LOG.info(
            "Performing OptaDOS pDIS calculation with parameters from {}".format(
                odi_fname
            )
        )
        return _run_optados(computer, odi_dict, seed, suffix="dispersion")

    return None


def _run_optados(computer, odi_dict, seed, suffix=None):
    """Run OptaDOS with given computer object, parameters and seed, adjusting
    the number of cores and the executable to call, then restoring them after.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling OptaDOS.
        odi_dict (dict): the OptaDOS parameters to write to file.
        seed (str): root filename of structure.

    Keyword arguments:
        suffix (str): either 'dos' or 'dispersion' for backup/restore.
        executable (str): optados executable to call.

    """

    from matador.export import doc2arbitrary

    odi_path = "{}.odi".format(seed)
    if computer.compute_dir is not None:
        odi_path = computer.compute_dir + "/" + odi_path
    doc2arbitrary(odi_dict, odi_path, overwrite=True)

    if suffix is not None:
        _get_correct_files_for_optados(seed, suffix=suffix)

    _cache_executable = copy.deepcopy(computer.executable)
    _cache_core = copy.deepcopy(computer.ncores)
    _cache_nodes = copy.deepcopy(computer.nnodes)
    computer.ncores = 1
    computer.nnodes = 1
    computer.executable = computer.optados_executable
    success = False

    try:
        success = computer.run_generic(intermediate=True, mv_bad_on_failure=False)
    except Exception as exc:
        LOG.warning("Failed to call optados with error: {}".format(exc))

    computer.ncores = _cache_core
    computer.nnodes = _cache_nodes
    computer.executable = _cache_executable
    if suffix is not None:
        _get_correct_files_for_optados(seed, suffix="bak")

    return success


def _get_optados_fname(seed):
    """Get the most likely OptaDOS input file name, which here means either
    only existing one, or the shortest one.

    Parameters:
        seed (str): seedname for the calculation.

    Returns:
        str: the OptaDOS input filename.

    """
    odi_fname = None
    if os.path.isfile(seed + ".odi"):
        os.remove(seed + ".odi")

    if glob.glob("*.odi"):
        # dodginess: choose the odi file with the shortest name...
        shortest_fname = None
        for fname in glob.glob("*.odi"):
            if shortest_fname is None or len(fname) < len(shortest_fname):
                shortest_fname = fname
        odi_fname = shortest_fname

    return odi_fname


def _get_correct_files_for_optados(seed, suffix=None):
    """If e.g. dispersion and DOS calculations were run previously, but
    it is unclear which exists as the current <seed>.<ext>, try to copy
    the old <seed>.<ext>_dispersion/dos files to the correct place.

    Parameters:
        seed (str): seedname for the calculation.

    Keyword arguments:
        suffix (str): either 'dos' or 'dispersion', or 'bak' to get old files back.

    """
    import shutil

    LOG.debug("Getting files for OptaDOS: {} {}".format(seed, suffix))
    if suffix is None:
        return
    files_to_cache = [
        ".pdos_bin",
        ".dome_bin",
        "-out.cell",
        ".bands",
        ".cell",
        ".param",
    ]
    for ext in files_to_cache:
        old_file = "{}{}_{}".format(seed, ext, suffix)
        current_file = "{}{}".format(seed, ext)
        if os.path.isfile(current_file):
            if suffix != "bak":
                backup_file = "{}{}_{}".format(seed, ext, "bak")
                shutil.copy2(current_file, backup_file)
            os.remove(current_file)
        if os.path.isfile(old_file):
            LOG.debug("Copying {} to {}".format(old_file, current_file))
            shutil.copy2(old_file, current_file)
            if suffix == "bak":
                os.remove(old_file)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The calculator module contains classes for use as DFT/atomistic
calculators within the compute module class.

"""


__all__ = ["CastepCalculator"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .castep import CastepCalculator



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule defines some useful chemical functions and
constants, with a focus on battery materials.

"""


import copy
import warnings
from typing import Dict, Tuple, Any, List

import numpy as np
from matador.data.constants import *  # noqa


EPS = 1e-8


def get_iupac_ordered_elements(elements: List[str]) -> List[str]:
    """Returns the list of elements in IUPAC order, i.e., that specified
    in Table VI of the IUPAC Red Book:

    Nomenclature of Inorganic Chemistry, IUPAC Recommendations 2005, IUPAC Red Book,
    RSC Publishing, 2005 [ISBN 0 85404 438 8]

    Returns:
        List of elements in the IUPAC order.

    """
    from matador.data.periodic_table import PERIODIC_TABLE

    return sorted(elements, key=lambda x: PERIODIC_TABLE[x].iupac_order)


def get_periodic_table():
    """Return some periodic table macros."""
    periodic_table = dict()
    periodic_table["I"] = ["Li", "Na", "K", "Rb", "Cs", "Fr"]
    periodic_table["II"] = ["Be", "Mg", "Ca", "Sr", "Ba", "Ra"]
    periodic_table["III"] = ["B", "Al", "Ga", "In", "Tl"]
    periodic_table["IV"] = ["C", "Si", "Ge", "Sn", "Pb"]
    periodic_table["V"] = ["N", "P", "As", "Sb", "Bi"]
    periodic_table["VI"] = ["O", "S", "Se", "Te", "Po"]
    periodic_table["VII"] = ["F", "Cl", "Br", "I", "At"]
    periodic_table["Tran"] = [
        "Sc",
        "Ti",
        "V",
        "Cr",
        "Mn",
        "Fe",
        "Co",
        "Ni",
        "Cu",
        "Zn",
        "Y",
        "Zr",
        "Nb",
        "Mo",
        "Tc",
        "Ru",
        "Rh",
        "Pd",
        "Ag",
        "Cd",
        "Hf",
        "Ta",
        "W",
        "Re",
        "Os",
        "Ir",
        "Pt",
        "Au",
        "Hg",
    ]
    periodic_table["Lan"] = [
        "La",
        "Ce",
        "Pr",
        "Nd",
        "Pm",
        "Sm",
        "Eu",
        "Gd",
        "Tb",
        "Dy",
        "Ho",
        "Er",
        "Tm",
        "Yb",
        "Lu",
    ]
    periodic_table["Act"] = [
        "Ac",
        "Th",
        "Pa",
        "U",
        "Np",
        "Pu",
        "Am",
        "Cm",
        "Bk",
        "Cf",
        "Es",
        "Fm",
        "Md",
        "No",
        "Lr",
    ]
    periodic_table["X"] = [
        elem for group in periodic_table for elem in periodic_table[group]
    ]
    return periodic_table


def get_molar_mass(elem):
    """Returns molar mass of chosen element."""
    from matador.data.periodic_table import PERIODIC_TABLE

    try:
        return PERIODIC_TABLE[elem].mass
    except KeyError as exc:
        raise KeyError(f"{elem} is not a valid element symbol.") from exc


def get_atomic_number(elem):
    """Returns atomic number of chosen element."""
    from matador.data.periodic_table import PERIODIC_TABLE

    try:
        return PERIODIC_TABLE[elem].number
    except KeyError as exc:
        raise KeyError(f"{elem} is not a valid element symbol.") from exc


def get_atomic_symbol(atomic_number):
    """Returns elemental symbol from atomic number."""
    from matador.data.periodic_table import PERIODIC_TABLE

    try:
        return list(PERIODIC_TABLE)[atomic_number]
    except ValueError as exc:
        raise KeyError(
            f"{atomic_number} not contained within periodic table from this package!"
        ) from exc


def get_concentration(doc, elements, include_end=False):
    """Returns x for A_x B_{1-x}
    or x,y for A_x B_y C_z, (x+y+z=1).

    Parameters:
        doc (list/dict): structure to evaluate OR matador-style stoichiometry.
        elements (list): list of element symbols to enforce ordering.

    Keyword arguments:
        include_end (bool): whether or not to return the final value, i.e.
            [x, y, z] rather than [x, y] in the above.

    Returns:
        list of float: concentrations of elements in given order.

    """
    if isinstance(doc, dict):
        if doc.get("stoichiometry") is None:
            raise RuntimeError("No stoichiometry found.")

        stoich = doc["stoichiometry"]
    else:
        stoich = doc

    concs = [0.0] * (len(elements) - bool(not include_end))
    for _, elem in enumerate(stoich):
        if (include_end and elem[0] in elements) or (
            not include_end and elem[0] in elements[:-1]
        ):
            concs[elements.index(elem[0])] = elem[1] / float(get_atoms_per_fu(doc))
    return concs


def get_subscripted_formula(formula: str) -> str:
    """Subscripts numbers in a chemical formula, returning unicode subscripts."""
    return formula.translate(str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉"))


def get_subscripted_formula_tex(formula: str) -> str:
    """Subscripts numbers in a chemical formula, returning LaTeX math subscripts."""
    return get_formula_from_stoich(get_stoich_from_formula(formula), tex=True)


def get_num_intercalated(cursor):
    """Return array of the number of intercalated atoms
    per host atom from a list of structures, of type defined by
    the first entry in the structures' concentration vectors.

    Parameters:
        cursor (list of dict): structures to evaluate.

    Returns:
        ndarray: number of intercalated ions in each structure.

    """
    from .cursor_utils import get_array_from_cursor

    x = np.zeros((len(cursor)))
    comps = get_array_from_cursor(cursor, "concentration")
    for idx, comp in enumerate(comps):
        if len(comp) > 1:
            comp = comp[0]
        if 1 - comp == 0:
            x[idx] = np.NaN
        else:
            x[idx] = comp / (1 - comp)
    return x


def get_binary_grav_capacities(x, m_B):
    """Returns capacity in mAh/g from x/y in A_x B_y
    and m_B in a.m.u.
    """
    x = np.array(x)
    if m_B != 0:
        return x * FARADAY_CONSTANT_Cpermol * Cperg_to_mAhperg / m_B
    return float("NaN")


def get_generic_grav_capacity(concs, elements):
    """Returns gravimetric capacity of
    <elements[0]> in mAh/g of matador doc.
    """
    tmp_concs = np.array(concs, copy=True)
    # if no Li, capacity = 0...
    # tmp_concs /= np.min(concs)
    x = tmp_concs[0]
    if x == 0:
        return 0.0
    masses = dict()
    m_B = 0
    for elem in elements:
        masses[elem] = get_molar_mass(elem)
    for ind, elem in enumerate(elements):
        if ind == 0:
            continue
        else:
            m_B += masses[elem] * tmp_concs[ind]
    Q = get_binary_grav_capacities(x, m_B)
    return Q


def get_binary_volumetric_capacity(initial_doc, final_doc):
    """For initial (delithiated/sodiated) (single element) structure
    and final (maximally charged) binary structure, calculate the volumetric capacity.

    Parameters:
        initial_doc (dict): matador doc of delithiated phase
        final_doc (dict): matador doc of maximally lithiated phase

    Returns:
       volumetric_capacity (float): capacity in mAh/cm^3.

    """

    assert len(initial_doc["stoichiometry"]) == 1
    assert len(final_doc["stoichiometry"]) == 2

    for _ion in final_doc["stoichiometry"]:
        if _ion[0] != initial_doc["stoichiometry"][0][0]:
            ion = _ion[0]

    for species in final_doc["stoichiometry"]:
        if species[0] == ion:
            num_ion = species[1]
        else:
            num_B = species[1]

    num_ions_per_initial_fu = num_ion / num_B
    volume_per_fu_cm3 = (
        initial_doc["cell_volume"]
        * ANGSTROM_CUBED_TO_CENTIMETRE_CUBED
        / initial_doc["num_fu"]
    )
    return (num_ions_per_initial_fu / volume_per_fu_cm3) * (
        ELECTRON_CHARGE * Cperg_to_mAhperg
    )


def get_atoms_per_fu(doc):
    """Calculate and return the number of atoms per formula unit.

    Parameters:
        doc (list/dict): structure to evaluate OR matador-style stoichiometry.

    """
    if "stoichiometry" in doc:
        return sum([elem[1] for elem in doc["stoichiometry"]])
    return sum([elem[1] for elem in doc])


def get_formation_energy(chempots, doc, energy_key="enthalpy_per_atom"):
    """From given chemical potentials, calculate the simplest
    formation energy per atom of the desired document.

    Note:
        recursive_get(doc, energy_key) MUST return an energy per atom for
        the target doc and the chemical potentials.

    Parameters:
        chempots (list of dict): list of chempot structures, must be unique.
        doc (dict): structure to evaluate.

    Keyword arguments:
        energy_key (str or list): name of energy field to use to calculate
            formation energy. Can use a list of keys/subkeys/indices to
            query nested dicts with `matador.utils.cursor_utils.recursive_get`.

    Returns:
        float: formation energy per atom.

    """
    from matador.utils.cursor_utils import recursive_get

    # warn user if per_atom energy is not found
    if (
        isinstance(energy_key, (list, tuple))
        and not any(["per_atom" in str(key) for key in energy_key])
        or (isinstance(energy_key, str) and "per_atom" not in energy_key)
    ):
        warnings.warn(
            "Requested energy key {} in get_formation_energy may"
            " not be per atom, if so results will be incorrect.".format(energy_key)
        )

    try:
        formation = recursive_get(doc, energy_key)
    except KeyError as exc:
        print("Doc {} missing key {}".format(doc.get("source"), energy_key))
        raise exc

    # see if num chempots has been set and try to reuse it
    if "num_chempots" in doc:
        num_chempots = doc["num_chempots"]
    else:
        num_chempots = get_number_of_chempots(doc, chempots)

    num_atoms_per_fu = get_atoms_per_fu(doc)
    for ind, mu in enumerate(chempots):
        num_atoms_per_mu = get_atoms_per_fu(mu)
        try:
            mu_energy = recursive_get(mu, energy_key)
        except KeyError as exc:
            raise exc(
                "Chemical potential {} missing key {}".format(mu["source"], energy_key)
            )

        formation -= mu_energy * num_chempots[ind] * num_atoms_per_mu / num_atoms_per_fu

    return formation


def get_number_of_chempots(stoich, chempot_stoichs, precision=5):
    """Return the required number of each (arbitrary) chemical potentials
    to construct one formula unit of the input stoichiometry. Uses least-squares
    as implemented by `numpy.linalg.lstsq` and rounds the output precision based
    on the `precision` kwarg.

    Parameters:
        stoich (list/dict): matador-style stoichiometry,
            e.g. [['Li', 3], ['P', 1]], or the full document.
        chempot_stoichs (list/dict): list of stoichiometries of the input
            chemical potentials, or the full documents.

    Keyword arguments:
        precision (int/None): number of decimal places to round answer to. None
            maintains the precision from `numpy.linalg.lstsq`.

    Returns:
        list: number of each chemical potential required to create
            1 formula unit.

    Raises:
        RuntimeError: if the stoichiometry provided cannot be created
            with the given chemical potentials.

    """

    # need to support dict, list and Crystal inputs
    try:
        stoich = stoich["stoichiometry"]
    except (TypeError, ValueError):
        pass
    try:
        chempot_stoichs = [mu["stoichiometry"] for mu in chempot_stoichs]
    except (TypeError, ValueError):
        pass

    # find all elements present in the chemical potentials
    elements = set()
    for mu in chempot_stoichs:
        for elem, _ in mu:
            elements.add(elem)
    elements = sorted(list(elements))

    chempot_matrix = np.asarray(
        [get_padded_composition(mu, elements) for mu in chempot_stoichs]
    )

    try:
        solution = np.asarray(get_padded_composition(stoich, elements))
    except RuntimeError:
        raise RuntimeError(
            "Stoichiometry {} could not be created from chemical potentials {}: missing chempot".format(
                stoich, chempot_stoichs
            )
        )

    try:
        num_chempots, residuals, _, _ = np.linalg.lstsq(
            chempot_matrix.T, solution, rcond=None
        )
    except np.linalg.LinAlgError:
        raise RuntimeError(
            "Stoichiometry {} could not be created from chemical potentials {}: numpy LinAlg error".format(
                stoich, chempot_stoichs
            )
        )

    # check if lstsq actually found a "solution"
    if np.abs(np.sum(residuals)) > EPS:
        raise RuntimeError(
            "Stoichiometry {} could not be created from chemical potentials {}".format(
                stoich, chempot_stoichs
            )
        )

    # round output array based on user-specified precision
    if precision is not None:
        num_chempots[np.where(np.abs(num_chempots) < EPS)] = 0.0
        for i, val in enumerate(num_chempots):
            if np.abs(val - round(val, precision)) < EPS:
                num_chempots[i] = round(val, precision)

    # check for sensible numbers in output
    if np.min(np.sign(num_chempots)) == -1:
        raise RuntimeError(
            "Stoichiometry {} could not be created from chemical potentials {}".format(
                stoich, chempot_stoichs
            )
        )

    return num_chempots.tolist()


def get_stoich(atom_types):
    """Return integer stoichiometry from atom_types list.

    Parameters:
        atom_types (list): list of element symbols of each atom.

    Returns:
        list: matador-style stoichiometry, e.g. [['Li', 1], ['P', 2]].

    """
    from collections import defaultdict
    from math import gcd

    stoich = defaultdict(float)
    for atom in atom_types:
        if atom not in stoich:
            stoich[atom] = 0
        stoich[atom] += 1
    gcd_val = 0
    for atom in atom_types:
        if gcd_val == 0:
            gcd_val = stoich[atom]
        else:
            gcd_val = gcd(stoich[atom], gcd_val)
    # convert stoichiometry to tuple for fryan
    temp_stoich = []
    try:
        for key, value in stoich.items():
            if float(value) / gcd_val % 1 != 0:
                temp_stoich.append([key, float(value) / gcd_val])
            else:
                temp_stoich.append([key, value / gcd_val])
    except AttributeError:
        for key, value in stoich.iteritems():
            if float(value) / gcd_val % 1 != 0:
                temp_stoich.append([key, float(value) / gcd_val])
            else:
                temp_stoich.append([key, value / gcd_val])
    return sorted(temp_stoich)


def get_padded_composition(stoichiometry, elements):
    """Return a list that contains how many of each species in
    elements exists in the given stoichiometry. e.g. for [['Li', 2], ['O', 1]]
    with elements ['O', 'Li', 'Ba'], this function will return [1, 2, 0].

    Parameters:
        stoichiometry (list): matador-style stoichiometry, as above.
        elements (list): order of element labels to pick out.

    """
    composition = []
    for element in elements:
        if not isinstance(element, str):
            raise RuntimeError("Found invalid element symbol {}".format(element))
        for species in stoichiometry:
            if not isinstance(species, list):
                raise RuntimeError(
                    "Found invalid stoichiometry {}".format(stoichiometry)
                )
            if species[0] == element:
                composition.append(species[1])
                break
            elif species[0] not in elements:
                raise RuntimeError(
                    "Extra element {} in stoichiometry".format(species[0])
                )
        else:
            composition.append(0)

    return composition


def get_ratios_from_stoichiometry(stoichiometry):
    """Get a dictionary of pairwise atomic ratios.

    Parameters:
        stoichiometry (list): matador-style stoichiometry.

    Returns:
        dict: dictionary of pairwise ratios, e.g. for K8SnP4,
            ratio_dict = {'KSn': 8, 'KP': 2, 'SnP': 0.25,
                          'SnK': 0.125, 'PSn': 4, 'PK': 0.5}.

    """
    ratio_dict = dict()
    for _, elem_i in enumerate(stoichiometry):
        for _, elem_j in enumerate(stoichiometry):
            if elem_j[0] != elem_i[0]:
                ratio_dict[elem_i[0] + elem_j[0]] = round(
                    float(elem_i[1]) / elem_j[1], 3
                )
    return ratio_dict


def get_stoich_from_formula(formula: str, sort=True):
    """Convert formula string, e.g. Li2TiP4 into a matador-style
    stoichiometry, e.g. [['Li', 2], ['Ti', 1], ['P', 4]].

    Parameters:
        formula (str): chemical formula of compound

    Returns:
        list: sorted matador-style stoichiometry.

    """
    from math import gcd
    import re

    parsed_elements = parse_element_string(formula, stoich=True)
    elements = []
    fraction = []
    for i, _ in enumerate(parsed_elements):
        if not bool(re.search(r"\d", parsed_elements[i])):
            elements.append(parsed_elements[i])
            try:
                fraction.append(float(parsed_elements[i + 1]))
            except (ValueError, IndexError):
                fraction.append(1.0)
    gcd_val = 0
    for frac in fraction:
        if gcd_val == 0:
            gcd_val = frac
        else:
            gcd_val = gcd(int(frac), int(gcd_val))
    fraction = np.asarray(fraction)
    fraction /= gcd_val
    stoich = [[elements[ind], fraction[ind]] for ind, _ in enumerate(elements)]
    if sort:
        return sorted(stoich)

    return stoich


def parse_element_string(elements_str, stoich=False):
    """Parse element query string with macros. Has to parse braces
    too, and throw an error if brackets are unmatched.

    e.g.
        Parameters: '[VII][Fe,Ru,Os][I]'
        Returns: ['[VII]', '[Fe,Ru,Os]', '[I]']

    e.g.2
        Parameters: '[VII]2[Fe,Ru,Os][I]'
        Returns: ['[VII]2', '[Fe,Ru,Os]', '[I]']

    Parameters:
        elements_str: str, chemical formula, including macros.

    Keyword arguments:
        stoich: bool, parse as a stoichiometry, i.e. check for numbers

    Raises:
        RuntimeError: if the composition contains unmatched brackets.

    Returns:
        list: split list of elements contained in input

    """
    import re

    valid = False
    for char in elements_str:
        if char not in ["[", "]", "{", "}", ",", ":"] and not char.isalnum():
            raise RuntimeError("Illegal character {} detected in query.".format(char))
    valid = False
    for char in elements_str:
        if char.isupper():
            valid = True
            break
    if not valid:
        raise RuntimeError(
            "Composition must contain at least one upper case character."
        )
    elements = [elem for elem in re.split(r"([A-Z][a-z]*)", elements_str) if elem]
    if stoich:
        tmp_stoich = elements
        for ind, strng in enumerate(elements):
            if not any(char.isdigit() for char in strng):
                tmp_stoich[ind] = [strng]
            else:
                tmp_stoich[ind] = [
                    elem for elem in re.split(r"([0-9]+)", strng) if elem
                ]
        elements = [item for sublist in tmp_stoich for item in sublist]
    # split macros
    while (
        "[" in elements
        or "{" in elements
        or "][" in elements
        or "}{" in elements
        or "]{" in elements
        or "}[" in elements
    ):
        tmp_stoich = list(elements)
        cleaned = True
        while cleaned:
            for ind, tmp in enumerate(tmp_stoich):
                if tmp == "][":
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, "[")
                    tmp_stoich.insert(ind, "]")
                    cleaned = True
                elif tmp == "}{":
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, "{")
                    tmp_stoich.insert(ind, "}")
                    cleaned = True
                elif tmp == "]{":
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, "{")
                    tmp_stoich.insert(ind, "]")
                    cleaned = True
                elif tmp == "}[":
                    del tmp_stoich[ind]
                    tmp_stoich.insert(ind, "[")
                    tmp_stoich.insert(ind, "}")
                    cleaned = True
                elif ind == len(tmp_stoich) - 1:
                    cleaned = False
        for ind, tmp in enumerate(tmp_stoich):
            if tmp == "[":
                end_bracket = False
                while not end_bracket:
                    if tmp_stoich[ind + 1] == "]":
                        end_bracket = True
                    tmp_stoich[ind] += tmp_stoich[ind + 1]
                    del tmp_stoich[ind + 1]
        for ind, tmp in enumerate(tmp_stoich):
            if tmp == "{":
                end_bracket = False
                while not end_bracket:
                    if tmp_stoich[ind + 1] == "}":
                        end_bracket = True
                    tmp_stoich[ind] += tmp_stoich[ind + 1]
                    del tmp_stoich[ind + 1]
        if "]" in tmp_stoich:
            tmp_stoich.remove("]")
        if "}" in tmp_stoich:
            tmp_stoich.remove("}")
        if "" in tmp_stoich:
            tmp_stoich.remove("")

        elements = tmp_stoich

    for elem in elements:
        if "}[" in elem or "]{" in elem:
            raise RuntimeError("Unmatched brackets in query string")

    return elements


def get_root_source(source):
    """Get the main file source from a doc's source list.

    Parameters:
        source (str/list/dict): contents of doc['source'] or the doc itself.

    Returns:
        str: "root" filename, e.g. if source = ['KP.cell', 'KP.param',
            'KP_specific_structure.res'] then root = 'KP_specific_structure'.

    """
    try:
        sources = copy.deepcopy(source["source"])
    except (KeyError, TypeError):
        sources = copy.deepcopy(source)

    if isinstance(source, str):
        return source

    src_list = set()
    for src in sources:
        if any(
            [
                src.endswith(ext)
                for ext in [
                    ".res",
                    ".castep",
                    ".history",
                    ".history.gz",
                    ".phonon",
                    ".phonon_dos",
                    ".bands",
                    ".cif",
                    ".magres",
                ]
            ]
        ):
            src_list.add(".".join(src.split("/")[-1].split(".")[0:-1]))
        elif "OQMD" in src.upper():
            src_list.add(src)
        elif "MP-" in src.upper():
            src_list.add(src)
        elif len(sources) == 1:
            src_list.add(src)
        elif src == "command_line":
            src_list.add("command line")

    if len(src_list) > 1:
        raise RuntimeError("Ambiguous root source {}".format(sources))
    if len(src_list) < 1:
        raise RuntimeError("Unable to find root source from {}".format(sources))

    return list(src_list)[0]


def get_formula_from_stoich(
    stoich, elements=None, tex=False, sort=True, unicode_sub=False, latex_sub_style=""
):
    """Get the chemical formula of a structure from
    its matador stoichiometry.

    Parameters:
        stoich (list): matador-style stoichiometry.

    Keyword arguments:
        elements (list): list of element symbols to enforce order.
        tex (bool): whether to print a LaTeX-compatibile string.
        unicode_sub (bool): use unicode subscripts
        latex_sub_style (str): a string to wrap subscripts in, e.g.
            r"\\mathrm" or r"\\text" (default is blank).

    Returns:
        str: the string representation of the chemical formula.

    """
    if tex and unicode_sub:
        raise RuntimeError(
            "Cannot produce a formula with both LaTeX and unicode subscripts."
        )

    form = ""
    if not isinstance(stoich, list):
        stoich = stoich.tolist()
    if sort:
        stoich = sorted(stoich)
    if elements is None:
        elements = get_iupac_ordered_elements([s[0] for s in stoich])

    if elements is not None:
        for targ_elem in elements:
            for elem in stoich:
                if elem[0] == targ_elem:
                    if elem[1] == 1:
                        form += elem[0]
                    elif int(elem[1]) != 0:
                        if tex:
                            if elem[1] % 1 == 0:
                                elem[1] = int(elem[1])
                            form += r"{}$_{}{{{}}}$".format(
                                elem[0], latex_sub_style, elem[1]
                            )
                        else:
                            form += elem[0] + str(int(elem[1]))
        assert form != ""

    if unicode_sub:
        form = get_subscripted_formula(form)

    return form


def magres_reference_shifts(
    magres: Dict[str, Any],
    reference: Dict[str, Tuple[float, float]],
):
    """Set chemical shifts inside a matador document from shieldings and a given reference.

     Parameters:
         magres: A matador document containing the structure and magres shielding data.
         reference: Reference values in the form `{element: [gradient, constant]}`.

    Returns:
         The input dictionary with the `chemical_shift_isos` key set to the referenced shifts.
    """

    chemical_shifts = magres.get(
        "chemical_shift_isos", len(magres["atom_types"]) * [None]
    )
    for ind, species in enumerate(magres["atom_types"]):
        if species in reference:
            chemical_shift = (
                magres["chemical_shielding_isos"][ind] * reference[species][0]
            ) + reference[species][1]
            chemical_shifts[ind] = chemical_shift

    magres["chemical_shift_isos"] = chemical_shifts



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some scraper functions for
NMR-related inputs and outputs, e.g. .magres files.

"""


from collections import defaultdict
from os import stat

import numpy as np
from matador.utils.cell_utils import cart2abc, cart2frac
from matador.utils.chem_utils import get_stoich
from matador.scrapers.utils import scraper_function, get_flines_extension_agnostic
from matador.data.constants import (
    ELECTRON_CHARGE,
    PLANCK_CONSTANT,
    BARN_TO_M2,
    EFG_AU_TO_SI,
)
from matador.data.magres import ELECTRIC_QUADRUPOLE_MOMENTS


@scraper_function
def magres2dict(fname, **kwargs):
    """Extract available information from .magres file. Assumes units of
    Angstrom and ppm for relevant quantities.
    """
    magres = defaultdict(list)
    flines, fname = get_flines_extension_agnostic(fname, "magres")
    magres["source"] = [fname]

    # grab file owner username
    try:
        from pwd import getpwuid

        magres["user"] = getpwuid(stat(fname).st_uid).pw_name
    except Exception:
        magres["user"] = "xxx"

    magres["magres_units"] = dict()
    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ["<atoms>", "[atoms]"]:
            i = 1
            while flines[line_no + i].strip().lower() not in ["</atoms>", "[/atoms]"]:
                split_line = flines[line_no + i].split()
                if not split_line:
                    i += 1
                    continue
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                if split_line[0] == "units":
                    magres["magres_units"][split_line[1]] = split_line[2]
                elif "lattice" in split_line:
                    lattice = split_line[1:]
                    for j in range(3):
                        magres["lattice_cart"].append(
                            [float(elem) for elem in lattice[j * 3 : (j + 1) * 3]]
                        )
                    magres["lattice_abc"] = cart2abc(magres["lattice_cart"])
                elif "atom" in split_line:
                    atom = split_line
                    magres["atom_types"].append(atom[1])
                    magres["positions_abs"].append([float(elem) for elem in atom[-3:]])
                i += 1
            break

    if "atom_types" in magres:
        magres["num_atoms"] = len(magres["atom_types"])
        magres["positions_frac"] = cart2frac(
            magres["lattice_cart"], magres["positions_abs"]
        )
        magres["stoichiometry"] = get_stoich(magres["atom_types"])

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ["<magres>", "[magres]"]:
            i = 1
            while flines[line_no + i].strip().lower() not in ["</magres>", "[/magres]"]:
                split_line = flines[line_no + i].split()
                if not split_line:
                    i += 1
                    continue
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                if split_line[0] == "units":
                    magres["magres_units"][split_line[1]] = split_line[2]
                elif "sus" in split_line:
                    magres["susceptibility_tensor"] = np.array(
                        [float(val) for val in split_line[1:]]
                    ).reshape(3, 3)

                elif "ms" in split_line:
                    ms = np.array([float(val) for val in split_line[-9:]]).reshape(3, 3)
                    s_iso = np.trace(ms) / 3

                    # find eigenvalues of symmetric part of shielding and order them to calc anisotropy eta
                    symmetric_shielding = _symmetrise_tensor(ms)
                    s_yy, s_xx, s_zz = _get_haeberlen_eigs(symmetric_shielding)
                    s_aniso = s_zz - (s_xx + s_yy) / 2.0
                    asymm = (s_yy - s_xx) / (s_zz - s_iso)

                    # convert from reduced anistropy to CSA
                    magres["magnetic_shielding_tensors"].append(ms)
                    magres["chemical_shielding_isos"].append(s_iso)
                    magres["chemical_shift_anisos"].append(s_aniso)
                    magres["chemical_shift_asymmetries"].append(asymm)

                elif "efg" in split_line:
                    efg = np.array([float(val) for val in split_line[-9:]]).reshape(
                        3, 3
                    )
                    species = split_line[1]

                    eigs = _get_haeberlen_eigs(efg)
                    v_zz, eta = eigs[2], (eigs[0] - eigs[1]) / eigs[2]

                    # calculate C_Q in MHz
                    quadrupole_moment = ELECTRIC_QUADRUPOLE_MOMENTS.get(species, 1.0)

                    C_Q = (
                        ELECTRON_CHARGE
                        * v_zz
                        * quadrupole_moment
                        * EFG_AU_TO_SI
                        * BARN_TO_M2
                    ) / (PLANCK_CONSTANT * 1e6)

                    magres["electric_field_gradients"].append(efg)
                    magres["quadrupolar_couplings"].append(C_Q)
                    magres["quadrupolar_asymmetries"].append(eta)

                i += 1

    for line_no, line in enumerate(flines):
        line = line.lower().strip()
        if line in ["<calculation>", "[calculation]"]:
            i = 1
            while flines[line_no + i].strip().lower() not in [
                "</calculation>",
                "[/calculation]",
            ]:
                if i > len(flines):
                    raise RuntimeError("Something went wrong in reader loop")
                # space important as it excludes other calc_code_x variables
                if "calc_code " in flines[line_no + i]:
                    magres["calculator"] = flines[line_no + i].split()[1]
                if "calc_code_version" in flines[line_no + i]:
                    magres["calculator_version"] = flines[line_no + i].split()[1]
                i += 1

    return dict(magres), True


def _symmetrise_tensor(t):
    """Returns the symmetrised tensor.

    Arguments:
        t: numpy array containing a NxN square tensor.

    Returns:
        np.ndarray: NxN numpy array containing the symmetric tensor.

    """
    return 0.5 * (t + t.T)


def _get_haeberlen_eigs(t: np.ndarray):
    """Return Haeberlen convention ordered eigenvalues of the passed tensor.

    ``|s_zz - s_iso| >= |s_xx - s_iso| >= |s_yy - s_iso|``

    Arguments:
        t: numpy array containing a NxN square tensor.

    Returns:
        np.ndarray: N-d numpy array containing the ordered eigenvalues.

    """
    eig_vals, eig_vecs = np.linalg.eig(t)
    eig_vals, eig_vecs = zip(
        *sorted(zip(eig_vals, eig_vecs), key=lambda eig: abs(eig[0] - np.trace(t) / 3))
    )

    return eig_vals



# coding: utf-8
# Distributed under the terms of the MIT License.

""" `matador` performs, aggregates, and analyses high-throughput
electronic structure calculations, primarily for crystal structure
prediction, leveraging CASTEP and Quantum Epsresso as density-functional
theory compute engines.
"""

import sys

__all__ = ["__version__"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"
__version__ = "0.10.1"

script_epilog = f"Written and maintained by Matthew Evans (me388@cam.ac.uk) 2016-2022, version {__version__}."


if sys.version_info.minor == 7:
    import warnings

    warnings.warn(
        "v0.10 of the `matador` library will be the last to support Python 3.7. "
        "Please upgrade to Python 3.8+ to use v0.11 and later versions of `matador`.",
        DeprecationWarning,
    )



# coding: utf-8
# Distributed under the terms of the MIT License.

import os
from glob import glob

from setuptools import find_packages, setup

from matador import __version__

with open("requirements/requirements.txt", "r") as f:
    requirements = [line.strip() for line in f.readlines()]

extra_requirements = dict(all=[])
req_files = glob("requirements/*.txt")
for _file in req_files:
    if _file != "requirements/requirements.txt":
        with open(_file, "r") as f:
            subreq = _file.split("/")[-1].split("_")[0]
            extra_requirements[subreq] = [line.strip() for line in f.readlines()]
            extra_requirements["all"] += extra_requirements[subreq]

with open("README.rst", "r") as f:
    long_description = f.read()


setup(
    name="matador-db",
    version=__version__,
    description="MATerial and Atomic Databases Of Refined structures.",
    long_description=long_description,
    url="https://github.com/ml-evs/matador",
    author="Matthew Evans",
    author_email="matador@ml-evs.science",
    maintainer="Matthew Evans",
    maintainer_email="matador@ml-evs.science",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=requirements,
    scripts=[script for script in glob("scripts/*") if os.path.isfile(script)],
    test_suite="tests",
    include_package_data=True,
    setup_requires=["setuptools>=42"],
    extras_require=extra_requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    entry_points={
        "console_scripts": [
            "matador = matador.cli.cli:main",
            "dispersion = matador.cli.dispersion:main",
            "run3 = matador.cli.run3:main",
        ]
    },
    zip_safe=False,
)



import unittest

try:
    import ase  # noqa

    ASE_IMPORTED = True
except ImportError:
    ASE_IMPORTED = False

try:
    import pymatgen  # noqa

    PMG_IMPORTED = True
except ImportError:
    PMG_IMPORTED = False

import numpy as np


from .utils import REAL_PATH


@unittest.skipIf((not (PMG_IMPORTED and ASE_IMPORTED)), "Unable to import ASE/pymtagen")
class PMGUtilTest(unittest.TestCase):
    """Tests cursor util functions."""

    def test_doc2pmg(self):
        from matador.scrapers import castep2dict

        doc, s = castep2dict(
            REAL_PATH + "data/castep_files/KP-castep17.castep", as_model=True
        )

        pmg_structure = doc.pmg_structure
        self.assertEqual(pmg_structure.formula, "K7 P2")
        np.testing.assert_array_almost_equal(
            pmg_structure.lattice.matrix, doc.lattice_cart, decimal=16
        )
        for ind, site in enumerate(doc):
            np.testing.assert_array_almost_equal(
                site.coords, pmg_structure[ind].frac_coords, decimal=16
            )
            self.assertEqual(site.species, pmg_structure[ind].species_string)

        self.assertDictEqual(pmg_structure.info["matador"], doc._data)

    def test_pmg2dict(self):
        import ase.build
        from pymatgen.io.ase import AseAtomsAdaptor
        from matador.utils.pmg_utils import pmg2dict

        # construct ASE atoms object manually
        ase_atoms = ase.build.bulk("SiO", crystalstructure="rocksalt", a=12, cubic=True)

        pmg_structure = AseAtomsAdaptor.get_structure(ase_atoms)
        doc = pmg2dict(pmg_structure, as_model=True)

        self.assertEqual(
            doc["atom_types"], sorted(["Si", "O", "Si", "O", "Si", "O", "Si", "O"])
        )

        np.testing.assert_array_almost_equal(
            pmg_structure.lattice.matrix, doc.lattice_cart, decimal=16
        )
        self.assertAlmostEqual(doc["cell_volume"], 12**3)
        self.assertEqual(len(doc["positions_frac"]), len(doc["atom_types"]))
        self.assertEqual(doc["num_atoms"], 8)
        self.assertEqual(doc["num_fu"], 4)
        self.assertListEqual(doc["stoichiometry"][0], ["O", 1.0])
        self.assertListEqual(doc["stoichiometry"][1], ["Si", 1.0])
        self.assertListEqual(sorted(list(doc["elems"])), ["O", "Si"])
        self.assertEqual(doc["space_group"], "Fm-3m")
        self.assertEqual(pmg_structure.formula, "Si4 O4")



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the Site class for handling
atomic sites.

"""

import numpy as np
from matador.utils.cell_utils import cart2frac, frac2cart, wrap_frac_coords
from matador.orm.orm import DataContainer


class Site(DataContainer):
    """The Site class contains a description of an individual
    site within a 3D periodic Crystal.

    """

    # This dictionary defines the map between fields in :obj:`Crystal`
    # that correspond to arrays of site properties and between the
    # relevant keys the :obj:`Site` object
    _crystal_key_map = {
        "site_occupancy": "site_occupancy",
        "chemical_shielding_isos": "chemical_shielding_iso",
        "chemical_shift_isos": "chemical_shift_iso",
        "magnetic_shielding_tensors": "magnetic_shielding_tensor",
        "electric_field_gradients": "electric_field_gradient",
        "chemical_shift_anisos": "chemical_shift_aniso",
        "chemical_shift_asymmetries": "chemical_shift_asymmetry",
        "quadrupolar_couplings": "quadrupolar_coupling",
        "quadrupolar_asymmetries": "quadrupolar_asymmetry",
        "voronoi_substructure": "voronoi_substructure",
    }

    def __init__(
        self,
        species: str,
        position: list,
        lattice,
        position_unit="fractional",
        mutable: bool = False,
        **site_data,
    ):
        """Initialise a Site object from its species, position and
        a reference to the lattice it exists in. Any other keys will be made available
        as site-level values.

        """
        if site_data.get("voronoi_substructure") is not None:
            assert self.species == site_data["voronoi_substructure"][0]
            site_data["voronoi_substructure"] = site_data["voronoi_substructure"][1]

        # DataContainer will take a copy of all data passed to it, but lets keep
        # lattice as a reference so that it can change externally
        self._lattice = lattice

        super().__init__(
            species=species, position=position, site_data=site_data, mutable=mutable
        )

        self._data["lattice_cart"] = self._lattice

        self.set_position(position, position_unit)
        self._occupancy = None

        self.site_data = {}
        self.site_data.update(site_data)

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, AttributeError):
            return default

    def __getitem__(self, key):
        """Add extra look-up in `self.site_data` to
        :class:`DataContainer`'s `__getitem__`.

        Parameters:
            key (str): name of key or attribute to get.

        Raises:
            AttributeError: if key or attribute can't be found.

        """
        if isinstance(key, int):
            raise ValueError("Object does not support indexing")
        try:
            super().__getitem__(key)
        except (AttributeError, KeyError):
            pass

        try:
            return self.site_data[key]
        except KeyError:
            raise KeyError(
                'Site has no data/site_data or implementation for requested key: "{}"'.format(
                    key
                )
            )

    def __setitem__(self, key: str, item):
        if key not in self.site_data or self.site_data[key] is None:
            self.site_data[key] = item
            return

        elif self.site_data[key] != item:
            try:
                import math

                if math.isnan(item) and math.isnan(self.site_data[key]):
                    return
            except TypeError:
                pass
            raise AttributeError(
                "Cannot assign value {} to existing key {} with value {}".format(
                    item, key, self.site_data[key]
                )
            )

    def __str__(self):
        site_str = "{species} {pos[0]:4.4f} {pos[1]:4.4f} {pos[2]:4.4f}".format(
            species=self.species, pos=self.coords
        )
        for key in self.site_data:
            try:
                site_str += "\n{} = {:4.4f}".format(
                    key, np.asarray(self.site_data[key])
                )
            except (ValueError, TypeError):
                with np.printoptions(precision=2, threshold=6, edgeitems=2):
                    site_str += "\n{} = \n{}".format(
                        key,
                        "\n".join(
                            f"  {row}"
                            for row in np.asarray(self.site_data[key])
                            .__str__()
                            .split("\n")
                        ),
                    )

        site_str += "\n---"

        return site_str

    def __deepcopy__(self, memo):
        from copy import deepcopy

        species, position, lattice = (
            deepcopy(x) for x in (self.species, self.coords, self.lattice)
        )
        site_data = deepcopy(self.site_data)
        return Site(species, position, lattice, position_unit="fractional", **site_data)

    def set_position(self, position, units):
        if len(position) != 3 or not all(isinstance(p, (float, int)) for p in position):
            raise RuntimeError(
                "CrystalSite position has wrong shape: {}".format(position)
            )
        if not hasattr(self, "_coords"):
            self._coords = dict()
        if units == "fractional":
            self._coords["fractional"] = wrap_frac_coords(
                [float(pos) for pos in position], remove=False
            )
        elif units == "cartesian":
            self._coords["fractional"] = wrap_frac_coords(
                cart2frac(self.lattice, self.coords), remove=False
            )
        else:
            raise RuntimeError(
                "Unit system {} not understood, expecting `fractional`/`cartesian`".format(
                    units
                )
            )

    @property
    def coords(self):
        return np.asarray(self._coords["fractional"])

    @property
    def coords_cartesian(self):
        return np.asarray(frac2cart(self.lattice, self.coords))

    @property
    def species(self):
        return self._data["species"]

    @species.setter
    def species(self, value):
        self._data["species"] = value

    @property
    def lattice(self):
        try:
            return self._lattice.lattice_cart
        except AttributeError:
            return self._lattice

    @property
    def occupancy(self):
        if "site_occupancy" not in self._data:
            self._data["site_occupancy"] = 1.0
        return self._data["site_occupancy"]

    @property
    def coordination(self):
        if "_coordination" in self.__dict__:
            return self._coordination
        if self.voronoi_substructure is None:
            raise RuntimeError("Voronoi substructure not found.")

        coordination = {}
        eps = 0.05
        for atom, weight in self.voronoi_substructure:
            if weight >= 1 - eps:
                if atom not in coordination:
                    coordination[atom] = 1
                else:
                    coordination[atom] += 1
        self._coordination = coordination

        return coordination

    def displacement_between_sites(self, other_site):
        return self.coords_cartesian - other_site.coords_cartesian

    def distance_between_sites(self, other_site):
        return np.linalg.norm(self.displacement_between_sites(other_site))
